from __future__ import division, print_function

import copy

import numpy as np
import networkx as nx

import matplotlib.pyplot as plt
import matplotlib.cm as cm
import operator
from sklearn.cluster import KMeans
from plots import multi_plot


class EigenvectorModularity(object):

    def __init__(self, g, heuristic_swap=False, null_weight=1.0):
        self.g = g
        self.dendrogram = None  #dendrogram
        self.is_directed = g.__class__.__name__ == 'DiGraph'
        self.B = None
        self.heuristic_swap = heuristic_swap
        self.null_weight = null_weight

    def compute_modularity_matrix(self, g, null_weight=1.0):
        """
            This method computes the modularity matrix from a graph g. Subclasses have to implement this.
        """
        raise NotImplementedError('Do not use EigenvectorModularity class directly, use one of it\'s subclasses!')

    def get_communities(self, min_size=None):
        """
            Main method, gets a list of communities from the graph specified in constructor. Uses the eigenvector
            method to do a greedy maximization of modularity.
        """

        if self.dendrogram is not None:
            communities = list()
            get_communities_from_dendrogram(self.dendrogram, communities, min_size=min_size)
            Qfinal = compute_modularity_from_matrix(self.g, self.B, communities)
            print('Final Modularity: %0.3f' % Qfinal)
            return communities

        #construct dendrogram that contains graph splits, leaves are communities
        dg = nx.DiGraph()
        dg_root_id = 'root'
        dg.add_node(dg_root_id)
        dg.node[dg_root_id]['subgraph_nodes'] = self.g.nodes()

        #compute modularity matrix (implemented by subclass)
        self.B = self.compute_modularity_matrix(self.g, null_weight=self.null_weight)
        assert self.B is not None
        assert self.B.shape == (self.g.number_of_nodes(), self.g.number_of_nodes())

        #find initial split
        sg1,sg2,Q = eigenvector_partition(self.g, self.B, self.is_directed, self.heuristic_swap)

        if len(sg1) == 0 or len(sg2) == 0:
            print('No initial partition found!')
            return [self.g.nodes()]

        #recursively find all the other partitions
        self.find_subcommunities(sg1, self.g, self.B, dg, dg_root_id, 'a')
        self.find_subcommunities(sg2, self.g, self.B, dg, dg_root_id, 'b')
        self.dendrogram = dg

        #build list of communities from dendrogram
        communities = list()
        get_communities_from_dendrogram(self.dendrogram, communities, min_size=min_size)
        Qfinal = compute_modularity_from_matrix(self.g, self.B, communities)
        print('Final Modularity: %0.3f' % Qfinal)

        return communities


    def find_subcommunities(self, subgraph_nodes, g, B, dg, dg_parent_id, handedness):
        """
            Recursive function for finding splits in subgraphs for the eigenvector modularity method.
        """
        Bg = compute_generalized_modularity_matrix(g, B, subgraph_nodes, is_directed=self.is_directed)
        sg = g.subgraph(subgraph_nodes)
        #prune_graph(sg)
        if len(sg) == 0:
            return

        #append to dendrogram
        #dg_sg_id = subgraph_md5(sg)
        dg_sg_id = '%s->%s' % (dg_parent_id, handedness)
        dg.add_node(dg_sg_id)
        dg.node[dg_sg_id]['subgraph_nodes'] = sg.nodes()
        dg.add_edge(dg_sg_id, dg_parent_id)

        sg1,sg2,Q = eigenvector_partition(sg, Bg, self.is_directed, heuristic_swap=self.heuristic_swap)
        if len(sg1) == 0 or len(sg2) == 0 or Q < 0.0:
            return

        #recursively find other partitions
        self.find_subcommunities(sg1, g, B, dg, dg_sg_id, 'a')
        self.find_subcommunities(sg2, g, B, dg, dg_sg_id, 'b')



class UndirectedModularity(EigenvectorModularity):
    """
        Implementation of eigenvector modularity maximization for undirected weighted graphs.
    """

    def __init__(self, g, heuristic_swap=False, null_weight=1.0):
        EigenvectorModularity.__init__(self, g, heuristic_swap=heuristic_swap, null_weight=null_weight)

    def compute_modularity_matrix(self, g, null_weight=1.0):

        if self.B is not None:
            return self.B

        #compute modularity matrix
        N = g.number_of_nodes()


        #compute total weight across all edges
        total_weights = dict()
        for n1 in g.nodes():
            tw = 0.0
            for n2 in g[n1]:
                tw += g[n1][n2]['weight']
            total_weights[n1] = tw

        M = np.array(total_weights.values()).sum()

        #compute modularity
        B = np.zeros([N, N])
        for i,ni in enumerate(g.nodes()):
            for j,nj in enumerate(g.nodes()):
                wi = total_weights[ni]
                wj = total_weights[nj]
                aij = 0.0
                if nj in g[ni]:
                    aij = g[ni][nj]['weight']
                B[i, j] = aij - null_weight*((wi*wj) / M)

        self.B = B
        return self.B



class LinkRankModularity(EigenvectorModularity):
    """
        Implementation of directed graph modularity maximization from:
        Youngdo Kim, Seung-Woo Son, and Hawoong Jeong,
         "Finding communities in directed networks" PHYSICAL REVIEW E 81, 016103 2010
    """

    def __init__(self, g, heuristic_swap=False, null_weight=1.0, alpha=0.85):
        self.alpha = alpha
        EigenvectorModularity.__init__(self, g, heuristic_swap=heuristic_swap, null_weight=null_weight)

    def compute_modularity_matrix(self, g, null_weight=1.0):

        if self.B is not None:
            return self.B

        self.N = g.number_of_nodes()

        #compute page rank of each node
        self.page_rank = nx.pagerank(g, alpha=self.alpha)

        #compute google and link rank matrix
        self.G, self.L = compute_link_rank_matrix(g, self.page_rank, alpha=self.alpha)

        #compute modularity matrix
        self.B, self.Bnull = compute_link_rank_modularity_matrix(g, self.page_rank, self.L, null_weight=null_weight)

        return self.B



def compute_modularity_from_matrix(g, B, communities):
    """
        Computes the modularity (a scalar) from a modularity matrix and a labeling of each node
         and it's community. g is a networkx graph, B is a modularity matrix, and communities is
         a list of lists, which each sub-list containing a list of nodes within that community.
    """

    #compute total weight
    total_weight = 0.0
    for n1,n2 in g.edges():
        total_weight += g[n1][n2]['weight']

    #make a dictionary of nodes to communities
    node2comm = dict()
    for k,comm in enumerate(communities):
        for n in comm:
            node2comm[n] = k

    #compute modularity
    modularity = 0.0
    for i,n1 in enumerate(g.nodes()):
        for j,n2 in enumerate(g.nodes()):
            if node2comm[n1] == node2comm[n2]:
                modularity += B[i, j]

    if not type(g).__name__ == 'DiGraph':
        modularity /= 2*total_weight

    return modularity

def merge_dendrogram_small_children(dg, node_id, min_size):
    """
        Merge child nodes in a dendrogram with their siblings if they're smaller than min_size.
    """
    child_nodes = dg.predecessors(node_id)

    needs_merge = False
    for cnode in child_nodes:
        if len(dg.node[cnode]['subgraph_nodes']) < min_size:
            needs_merge = True
    if needs_merge:
        #break link with children, effectively leaving only the parent node in the dendrogram
        for cnode in child_nodes:
            dg.remove_edge(cnode, node_id)
    else:
        for cnode in child_nodes:
            merge_dendrogram_small_children(dg, cnode, min_size)

def iterate_dendrogram(dg, node_id, ifunc):
    ifunc(dg, node_id)
    for cnode in dg.predecessors(node_id):
        iterate_dendrogram(dg, cnode, ifunc)


def get_communities_from_dendrogram(dg, communities, min_size=None, node_id='root'):
    """ From a dendrogram, get communities. Communities are the leaves of the dendrogram, or if min_size is
        specified, the deepest one can go in the dendrogram without finding a community smaller than min_size.
        This is a recursive function, communities should be passed in as an empty array and will be filled on return.
    """

    if min_size is not None and node_id == 'root':
        #merge subgraphs with their siblings when they are smaller than min_size
        dg_copy = copy.copy(dg)
        merge_dendrogram_small_children(dg_copy, 'root', min_size)
        dg_to_use = dg_copy
    else:
        dg_to_use = dg

    nodes = dg_to_use.node[node_id]['subgraph_nodes']

    node_children = dg_to_use.predecessors(node_id)
    if len(node_children) == 0:
        #we're in a leaf node
        communities.append(nodes)
    else:
        for child in node_children:
            get_communities_from_dendrogram(dg_to_use, communities, min_size=min_size, node_id=child)


def eigenvector_partition(g, B, is_directed=False, heuristic_swap=False):
    """ Partition a graph into two subgraphs based on the eigenvector modularity method """

    #compute eigenvalues/eigenvectors, find max

    Bused = B
    if is_directed:
        #for directed graphs, there's a "trick". we want to find the eigenvalues of a symmetric matrix, so that
        #there are no complex-valued eigenvalues to screw up the method. we do that by adding the transpose of B
        Bused += B.transpose()

    U,V = np.linalg.eig(Bused) #find eigendecomposition
    kmax = U.argmax()
    umax = U[kmax] #largest eigenvalue
    vmax = V[:, kmax].squeeze() #eigenvector associated with largest eigenvalue

    assert np.imag(umax) == 0.0 #the maximum eigenvalue should not be complex (none of them should!)

    #if all eigenvalues are negative or zero, we cannot partition
    if umax <= 0.0:
        return [],[],0.0

    #find optimal partitioning
    s = np.sign(vmax)
    g1 = list()
    g2 = list()
    for k,n in enumerate(g.nodes()):
        if s[k] > 0.0:
            g1.append(n)
        else:
            g2.append(n)

    #compute modularity
    Q = np.dot(s, np.dot(Bused, s)) / (4*g.number_of_edges())

    if heuristic_swap:
        #swap the community of each node to try and further maximize modularity in a fine-tuning process
        print('Heuristic Swap turned on, starting Q=%0.6f' % Q)
        Qbest = Q
        Qdiff = 0.0
        sbest = s
        nodes_swapped = dict()
        start = True
        while start or Qdiff > 0.0:
            start = False
            Qdiff = 0.0

            #go through each node
            for k,n in enumerate(g.nodes()):

                #if the node has been successfully swapped already, ignore it
                if n in nodes_swapped:
                    continue

                #flip the community of node n
                snew = copy.copy(sbest)
                snew[k] = -sbest[k]
                #recompute the modularity
                Qnew = np.dot(snew, np.dot(Bused, snew)) / (4*g.number_of_edges())
                if Qnew > Qbest:
                    print('swapped node %d, Qold=%0.6f, Qnew=%0.6f' % (n, Qbest, Qnew))
                    #if we do better on modularity, keep this swap
                    Qdiff = Qbest - Qnew
                    Qbest = Qnew
                    sbest = snew
                    nodes_swapped[n] = True

        #recompute the subgraphs
        Q = Qbest
        g1 = list()
        g2 = list()
        for k,n in enumerate(g.nodes()):
            if sbest[k] > 0.0:
                g1.append(n)
            else:
                g2.append(n)

    return g1,g2,Q


def compute_generalized_modularity_matrix(g_parent, B, subgraph_nodes, is_directed=False):

    nodes_parent = g_parent.nodes()
    Ng = len(subgraph_nodes)
    Bg = np.zeros([Ng, Ng])

    for i in range(Ng):
        for j in range(Ng):
            ni = subgraph_nodes[i]
            nj = subgraph_nodes[j]
            i_parent = nodes_parent.index(ni)
            j_parent = nodes_parent.index(nj)
            bij = B[i_parent, j_parent]
            if i == j:
                dsum = 0.0
                for k in range(Ng):
                    k_parent = nodes_parent.index(subgraph_nodes[k])
                    dsum += B[i_parent, k_parent]
                    if is_directed:
                        dsum += B[k_parent, i_parent]
                m = 1.0 - (is_directed*0.5)  #multiplier is 1/2 for directed graphs, 1 for undirected
                bij -= m*dsum
            Bg[i, j] = bij
    return Bg


def compute_link_rank_modularity_matrix(g, page_rank, L, null_weight=1.0):

    N = g.number_of_nodes()
    Bnull = np.zeros([N, N], dtype='float')

    for i,n1 in enumerate(g.nodes()):
        for j,n2 in enumerate(g.nodes()):
            Bnull[i, j] = page_rank[n1]*page_rank[n2]

    B = L - null_weight*Bnull
    return B, Bnull


def compute_google_matrix(g, alpha=0.85):

    N = g.number_of_nodes()
    G = np.zeros([N, N], dtype='float')

    #compute total out weight for each node
    w_out = {}
    for n1 in g.nodes():
        w_out[n1] = 0.0
        for n2,eparams in g[n1].iteritems():
            w_out[n1] += eparams['weight']

    all_edges = g.edges()
    #compute google matrix
    for i,n1 in enumerate(g.nodes()):
        for j,n2 in enumerate(g.nodes()):
            if (n1,n2) in all_edges:
                ai = 1 - int(w_out[n1] > 0.0)
                gij = alpha*(g[n1][n2]['weight'] / w_out[n1]) + (1.0/N)*(alpha*ai + 1 - alpha)
                G[i, j] = gij

    return G


def compute_link_rank_matrix(g, page_rank, alpha=0.85):

    N = g.number_of_nodes()
    G = compute_google_matrix(g, alpha=alpha)
    L = np.zeros([N, N], dtype='float')

    for i,n1 in enumerate(g.nodes()):
        for j,n2 in enumerate(g.nodes()):
            L[i, j] = page_rank[n1] * G[i, j]

    return G, L

def prune_graph(g):
    """ Remove unconnected nodes """
    nodes2remove = []
    for n in g.nodes():
        if len(g[n]) == 0:
            nodes2remove.append(n)
    for n in nodes2remove:
        g.remove_node(n)


class SpectralCluster(object):
    """ Spectral cluster graph g (Azran and Ghahramani 2010) based on the affinity matrix P """

    def __init__(self, g, K=None, Mvals=range(1, 500), weight_param='weight'):
        self.g = g
        self.K = K
        self.Mvals = Mvals
        self.W,self.D,self.P = compute_affinity(self.g, weight_param=weight_param)
        self.compute_spectrum()

    def compute_principle_component(self, v):
        denom = np.dot(v, np.dot(self.D, v))
        return np.dot(np.outer(v, v), self.D) / denom

    def compute_spectrum(self):

        #compute spectrum of P
        eigenvalues,V = np.linalg.eig(self.P)

        #sort the spectrum from highest to lowest eigenvalue
        spectrum = list()
        for k,n in enumerate(self.g.nodes()):
            v = np.real(V[:, k])
            principle_component = self.compute_principle_component(v)
            spectrum.append( (n, np.real(eigenvalues[k]), v, principle_component, np.abs(np.real(eigenvalues[k]))) )
        spectrum.sort(key=operator.itemgetter(-1), reverse=True)
        self.spectrum = spectrum

        #compute the max gap \Delta{M}(M) and max gap index K(M) for each M
        ev = np.array([x[1] for x in spectrum])
        max_gap = np.zeros([len(self.Mvals)])
        max_gap_index = np.zeros([len(self.Mvals)])

        for k,M in enumerate(self.Mvals):
            pspec = ev**M
            pspec_diff = np.array([pspec[j] - pspec[j+1] for j in range(len(ev)-1)])
            max_gap_index[k] = pspec_diff.argmax() + 1
            max_gap[k] = pspec_diff.max()

        #determine the maximum value of M that makes sense
        one_gi = max_gap_index == 1
        if one_gi.sum() > 0:
            Mmax_index = np.where(one_gi)[0].min()
        else:
            Mmax_index = 0
        Mmax = self.Mvals[Mmax_index]

        self.M = np.array(range(1, Mmax+1))
        self.Mmax = Mmax
        self.max_gap = max_gap[:Mmax_index+1]
        self.max_gap_index = max_gap_index[:Mmax_index+1]

    def cluster(self, K=None):
        if K is not None:
            self.K = K

        #determine K if needed
        if self.K is None:
            raise NotImplementedError('Automatic determination of K not implemented!')

        #use K-means clustering on the eigenvectors
        N = len(self.spectrum)
        X = np.zeros([N, self.K-1])
        for k in range(self.K-1):
            eigenvector = self.spectrum[k+1][2]
            X[:, k] = eigenvector

        kmean = KMeans(self.K, init='k-means++', n_init=100, max_iter=500, precompute_distances=True)
        kmean.fit(X)

        #determine cluster membership from results of K-means
        clusters = kmean.predict(X)
        node_clusters = dict()
        unique_clusters = np.unique(clusters)
        for uc in unique_clusters:
            node_clusters[uc] = list()
        for k,cluster in enumerate(clusters):
            n = self.spectrum[k][0]
            node_clusters[cluster].append(n)

        self.node_clusters = node_clusters

        return node_clusters

    def plot(self, markersize=12):

        V = np.array([x[2] for x in self.spectrum]).transpose()
        plt.figure()
        plt.imshow(V, interpolation='nearest', aspect='auto')
        plt.xlabel('Eigenvector')
        plt.ylabel('Node')

        plt.figure()
        plt.imshow(V[:, :self.K], interpolation='nearest', aspect='auto')
        plt.xlabel('Eigenvector')
        plt.ylabel('Node')
        plt.xticks([])
        plt.xticks(range(self.K))

        ev = np.abs(np.array([x[1] for x in self.spectrum]))
        plt.figure()
        plt.subplot(3, 1, 1)
        plt.plot(ev, 'go', markersize=markersize)
        plt.ylabel('Eigenvalue')
        plt.title('M=1')
        plt.ylim(0.0, 1.0+0.1)
        plt.subplot(3, 1, 2)
        plt.plot(ev**2, 'go', markersize=markersize)
        plt.ylabel('Eigenvalue')
        plt.ylim(0.0, 1.0+0.1)
        plt.title('M=2')
        plt.subplot(3, 1, 3)
        plt.plot(ev**self.Mmax, 'go', markersize=markersize)
        plt.ylabel('Eigenvalue')
        plt.title('M=%d' % self.Mmax)
        plt.ylim(0.0, 1.0+0.1)

        plt.figure()
        plt.subplot(2, 1, 1)
        plt.plot(self.M, self.max_gap, 'k-')
        plt.xlabel('M')
        plt.ylabel('$\Delta{M}(M)$')
        plt.subplot(2, 1, 2)
        plt.plot(self.M, self.max_gap_index, 'k-')
        plt.xlabel('M')
        plt.ylabel('K(M)')

        dlist = list()
        for k in range(self.K+1):
            n,eigenvalue,eigenvector,principle_component,abs_eigenvalue = self.spectrum[k]
            dlist.append({'eigenvalue':eigenvalue, 'T':principle_component, 'k':k})
        multi_plot(dlist, plot_principle_component, nrows=3, ncols=3)


def plot_principle_component(pdata, ax):
    T = pdata['T']
    v = pdata['eigenvalue']
    k = pdata['k']
    plt.imshow(T, interpolation='nearest', aspect='auto')
    plt.title('$\lambda_%d$=%0.6f' % (k,v))
    plt.xticks([])
    plt.yticks([])


def get_weight_matrix(g, weight_param='weight'):
    N = len(g.nodes())
    W = np.zeros([N, N])
    node2index = dict()
    for k,n in enumerate(g.nodes()):
        node2index[n] = k
    for n1,n2 in g.edges():
        i = node2index[n1]
        j = node2index[n2]
        w = g[n1][n2][weight_param]
        W[i][j] = w
        W[j][i] = w
    return W


def compute_affinity(g, weight_param='weight'):
    """ Computes the affinity matrix for an undirected graph g. """

    #compute the matrix D = diag(d(1), ..., d(N)), where d(i) is the sum of weights for node i
    N = len(g.nodes())
    W = get_weight_matrix(g, weight_param=weight_param)

    node_weights = np.zeros([N])
    for k,n in enumerate(g.nodes()):
        node_weights[k] = W[k, :].sum()
    znodes = (node_weights == 0.0)
    if znodes.sum() > 0:
        print(node_weights)
        raise ValueError('Cannot compute affinity matrix for graph that has disconnected nodes, or whose weight sum is zero.')

    D = np.diag(node_weights)
    Dinv = np.linalg.inv(D)

    P = np.dot(Dinv, W)
    return W,D,P








from __future__ import division, print_function

import copy
import numpy as np
from soundsig.incrowd import ConvolutionalInCrowdModel, fast_conv, fast_conv_grouped


class ConvolutionalLinearModel(ConvolutionalInCrowdModel):
    """
        Wrapper class for use with ThresholdGradientDescent optimization class, also
        provides a gradient calculation.
    """

    def __init__(self, input, target, ndelays=None, bias=0.0, group_index=None, lags=None):
        ConvolutionalInCrowdModel.__init__(self, input, target, ndelays, bias, group_index, lags)

    def error_vector(self, x):
        bias = x[-1]
        filt = self.get_filter(x[:-1])
        if self.group_index is None:
            yhat = fast_conv(self.input, filt, self.time_lags, bias=bias)
        else:
            yhat = fast_conv_grouped(self.input, filt, self.time_lags, group_index=self.group_index, bias=bias)

        return self.target_val - yhat

    def error(self, x):
        esq = self.error_vector(x)**2
        return esq.sum()

    def grad(self, x):
        e = self.error_vector(x)

        gfilt = np.zeros([self.num_channels, self.ndelays])
        tlen = len(e)
        for k,lag in enumerate(self.time_lags):
            if lag > 0:
                #causal filter component
                stim_start_index = 0
                stim_end_index = tlen - lag
                r_start_index = lag
                r_end_index = tlen
            else:
                #acausal filter component
                stim_start_index = abs(lag)
                stim_end_index = tlen
                r_start_index = 0
                r_end_index = tlen - abs(lag)

            #print 'lag=%d, stim_start=%d, stim_end=%d, r_start=%d, r_end=%d' % (lag, stim_start_index, stim_end_index, r_start_index, r_end_index)

            gfilt[:, k] = (self.input.T[:, stim_start_index:stim_end_index] * e[r_start_index:r_end_index]).sum(axis=1)

        #print 'gfilt=',gfilt.ravel()

        g = np.zeros([self.ndelays*self.num_channels + 1])
        g[:-1] = -2*gfilt.ravel()
        g[-1] = -2*e.sum()

        return g


class ThresholdGradientDescent(object):

    def __init__(self, model, step_size=1e-3, threshold=1.0, earlystop_model=None, gradient_norm=True, group_indices=None, slope_threshold=-1e-3):
        self.threshold = threshold
        self.model = model
        self.step_size = step_size
        self.earlystop_model = earlystop_model
        self.errors = list()
        self.es_errors = list()
        self.params = model.params
        self.gradient_norm = gradient_norm
        self.converged = False
        self.num_iters_for_slope = 5
        self.slope = -np.Inf
        self.slope_threshold = slope_threshold
        self.iter = 0
        self.group_indices = group_indices
        self.groups = None
        self.best_params = None
        self.best_err = np.inf

        if self.group_indices is not None:
            self.groups = np.unique(self.group_indices)
            if len(threshold) > 1 and len(threshold) != len(self.groups):
                raise Exception('Number of thresholds specified must equal the number of unique groups!')
            self.group_indices_map = dict()
            for g in self.groups:
                self.group_indices_map[g] = np.where(self.group_indices == g)

    def iterate(self):

        #compute gradient, update parameters
        g = self.model.grad(self.params)

        #threshold out elements of the gradient
        if not np.isscalar(self.threshold):
            for k,(group,gindex) in enumerate(self.group_indices_map.iteritems()):
                gsub = g[gindex]
                if self.gradient_norm:
                    gnorm = np.linalg.norm(gsub)
                    if gnorm > 0.0:
                        gsub /= gnorm
                gabs = np.abs(gsub)
                gthresh = gabs.max()*self.threshold[k]
                gsub[gabs < gthresh] = 0.0
                g[gindex] = gsub

        else:
            if self.gradient_norm:
                gnorm = np.linalg.norm(g)
                if gnorm > 0.0:
                    g /= gnorm
            gabs = np.abs(g)
            gthresh = gabs.max()*self.threshold
            g[gabs < gthresh] = 0.0

        if np.abs(g).sum() == 0:
            self.converged = True
            if self.iter == 0:
                self.best_params = copy.copy(self.params)
        else:
            self.params = self.params - self.step_size*g

        #compute error, check for convergenc
        e = self.model.error(self.params)
        self.errors.append(e)

        if self.earlystop_model is not None:
            es_err = self.earlystop_model.error(self.params)
            self.es_errors.append(es_err)

            if es_err < self.best_err:
                self.best_err = es_err
                self.best_params = copy.copy(self.params)

            if len(self.es_errors) >= self.num_iters_for_slope:
                slope,intercept = np.polyfit(range(self.num_iters_for_slope), self.es_errors[-self.num_iters_for_slope:], 1)
                slope /= np.abs(np.array(self.es_errors[-self.num_iters_for_slope:]).mean())
                self.slope = slope
                if self.slope > self.slope_threshold:
                    self.converged = True
        else:
            if len(self.errors) >= self.num_iters_for_slope:
                slope,intercept = np.polyfit(range(self.num_iters_for_slope), self.errors[-self.num_iters_for_slope:], 1)
                slope /= np.abs(np.array(self.errors[-self.num_iters_for_slope:]).mean())
                self.slope = slope
                if self.slope > self.slope_threshold:
                    print('[ThresholdGradientDescent] cross slope threshold on iteration %d, slope=%f' %
                          (self.iter+1, self.slope))
                    self.converged = True

                if e < self.best_err:
                    self.best_err = e
                self.best_params = copy.copy(self.params)

        self.iter += 1


def finite_diff_grad(errorfunc, params, eps=1e-8):

    base_err = errorfunc(params)
    fdgrad = np.zeros(len(params))

    for k in range(len(params)):
        dparams = copy.deepcopy(params)
        dparams[k] += eps
        merr = errorfunc(dparams)
        fdgrad[k] = (merr - base_err) / eps

    return fdgrad


def find_index(arr, vals):
    """
        Find the elements of arr that take on values in vals. Returns a logical array.
    """
    index = np.zeros([len(arr)], dtype='bool')

    for v in vals:
        index |= arr == v

    return index



from __future__ import division, print_function

import copy
import numpy as np
import operator
import matplotlib.pyplot as plt

try:
    import spams
except:
    print("Cannot import spams! You won't be able to optimize!")

try:
    from cvxopt import matrix as cvxopt_matrix, solvers as cvxopt_solvers
except:
    print('Cannot import cvxopt, you won\'t be able to use the quadratic dense solver!')

import time

class SPAMSLassoSolver(object):

    def __init__(self, solver_params):

        self.lambda1 = 1.0
        if 'lambda1' in solver_params:
            self.lambda1 = solver_params['lambda1']

        self.lambda2 = 1.0
        if 'lambda2' in solver_params:
            self.lambda2 = solver_params['lambda2']

        self.positive = False
        if 'positive' in solver_params:
            self.positive = solver_params['positive']

    def solve(self, A, y, x0, as_signs):

        #print('dense_solver: y.shape=',y.shape)
        #print('dense_solver: A.shape=',A.shape)
        fy = np.asfortranarray(y.reshape(len(y), 1))
        fA = np.asfortranarray(A)
        #print('fy.shape=',fy.shape)
        #print('fA.shape=',fA.shape)
        if not self.positive:
            xnew = spams.lasso(fy, fA, mode=2, lambda1=self.lambda1, lambda2=self.lambda2)
        else:
            W = np.ones(len(x0))
            params = {'lambda1':self.lambda1, 'pos':True}
            xnew = spams.lassoWeighted(fy, fA, W, **params)

        xnew = np.array(xnew.todense()).reshape(x0.shape)

        #print('dense_solver: xnew.shape=',xnew.shape)

        return xnew

class CVXOPTQPSolver(object):
    def __init__(self, solver_params):
        self.lambda_val = 200.0
        if 'lambda' in solver_params:
            self.lambda_val = solver_params['lambda']

    def solve(self, A, y, x0, as_signs):

        print('cvxopt QP: # of params=%d' % A.shape[1])
        #flip sign of columns of A that have negative usefulness, a trick
        #Patrick Gill uses to halve the # of parameters for the QP solver
        A *= as_signs
        #the sign of x0 is unflipped for the nonzero elements, flip it
        x0 *= as_signs

        p = A.shape[1]
        H = cvxopt_matrix(np.dot(A.transpose(), A))
        f = cvxopt_matrix(self.lambda_val - np.dot(A.transpose(), y))
        b = cvxopt_matrix(np.zeros(p))
        Q = cvxopt_matrix(-np.eye(p))

        stime = time.time()
        sol = cvxopt_solvers.qp(H, f, Q, b, None, None, None, x0)
        etime = time.time() - stime
        print('QP solver took %d seconds' % int(etime))

        xnew = np.array(sol['x']).squeeze()
        #unflip elements of xnew
        xnew *= as_signs

        return xnew


class InCrowd(object):
    """ Implementation of in-crowd basis pursuit denoiser. For reference see:
        "The In-Crowd Algorithm for Fast Basis Pursuit Denoising"
        Patrick R. Gill, Albert Wang, Alyosha Molnar
        IEEE TRANSACTIONS ON SIGNAL PROCESSING, VOL. 59, NO. 10, OCTOBER 2011
    """

    def __init__(self, model, solver_class=SPAMSLassoSolver, solver_params=dict(), max_contrast_ratio=np.Inf, max_additions_fraction=0.75):
        self.model = model

        self.max_contrast_ratio = max_contrast_ratio
        self.max_additions_fraction = max_additions_fraction
        self.dense_solver = solver_class(solver_params)

        self.x = self.model.initial_guess()

        self.active_set = []
        self.inactive_set = np.arange(0.0, len(self.x), 1.0, dtype='int').tolist()

        self.converged = False
        self.num_iter = 0
        self.last_active_set = None
        self.active_set_signs = {}

    def iterate(self):

        print('Active set size: %d' % len(self.active_set))
        if len(self.active_set) == 0 and self.num_iter > 0:
            print('Empty active set, converging!')
            self.converged = True
            return

        if len(self.inactive_set) > 0:
            #compute usefulness
            u = self.model.usefulness(self.x, self.inactive_set)
            #normalize usefulness
            uabsmax = np.abs(u).max()
            if uabsmax > 0.0:
                u /= uabsmax

            #sort usefulness from high to low
            ulist = [(self.inactive_set[k], np.abs(uval), uval) for k,uval in enumerate(u)]
            ulist.sort(key=operator.itemgetter(1), reverse=True)
            #for (ui, uabs, u) in ulist:
            #    print( (ui, uabs, u))
            ulist = np.array(ulist)

            #get most useful elements to add to active set by thresholding using max_additions_fraction
            nzi = (ulist[:, 1] >= self.max_additions_fraction).nonzero()[0]
            #num_to_use = min(len(nzi), 25)
            num_to_use = len(nzi)
            nzi = nzi[:num_to_use]
            new_active_indices = ulist[nzi, 0].astype('int')
            if len(nzi) > 0:
                print('Adding %d active elements' % len(nzi))
            #print('new_active_indices=',new_active_indices)

            #record signs of new active set elements
            for nz in nzi:
                usign = np.sign(ulist[nz, 2])
                uindex = ulist[nz, 0]
                self.active_set_signs[uindex] = usign

            #remove newly active parameters from inactive set, add them to active set
            new_inactive_set = copy.deepcopy(self.inactive_set)

            for x_index in new_active_indices:
                new_inactive_set.remove(x_index)
                self.active_set.append(x_index)
            self.inactive_set = new_inactive_set
            #print('Active set size: %d' % len(self.active_set))

        #print('active_set=',self.active_set)
        #print('inactive_set=',self.inactive_set)

        #get submatrix of active components
        Asub = self.model.submatrix(self.active_set, include_bias=True)
        ysub = self.model.target()
        x0sub = np.concatenate([self.x[self.active_set], [self.model.bias]])

        #print('Asub.shape=',Asub.shape)
        #print('ysub.shape=',ysub.shape)
        #print('x0sub.shape=',x0sub.shape)

        #construct array of usefulness signs, used by the QP dense solver
        as_signs = []
        for asindex in self.active_set:
            as_signs.append(self.active_set_signs[asindex])
        as_signs.append(1.0)  #for the bias term

        #use interior solver to get solution for new active set
        xsub = self.dense_solver.solve(Asub, ysub, x0sub, np.array(as_signs))
        self.model.bias = xsub[-1]
        xsub = xsub[:-1]

        #print('type(xsub)=',type(xsub).__name__)

        #threshold and set new x
        self.threshold_xsub(xsub)
        #print('thresholded xsub=',xsub)
        self.x[self.active_set] = xsub

        #kick zeros out of active set
        zi = (xsub == 0.0).nonzero()[0]
        new_active_set = copy.copy(self.active_set)
        for ip in zi:
            actual_index = self.active_set[ip]
            self.inactive_set.append(actual_index)
            new_active_set.remove(actual_index)
            del self.active_set_signs[actual_index]

        if len(zi) > 0:
            print('Removing %d elements' % len(zi))

        self.active_set = new_active_set

        #check for convergence
        if self.max_additions_fraction == 0.0:
            print('Threshold is 1.0, converging.')
            self.converged = True
        else:
            if self.num_iter > 0 and len(self.active_set) == len(self.last_active_set):
                if len(np.setdiff1d(self.active_set, self.last_active_set)) == 0:
                    print('Active set has not changed, converging!')
                    self.converged = True

        self.num_iter += 1
        self.last_active_set = copy.copy(self.active_set)

    def threshold_xsub(self, xsub):
        xsubabs = np.abs(xsub)
        zero_cutoff = xsubabs.max() / self.max_contrast_ratio
        xsub[xsubabs < zero_cutoff] = 0.0


class InCrowdModel(object):
    """ Represents a problem where we want to minimize the expression
        ||y-Ax||_2 - lambda*||x||_1, where ||.||_2 and ||.||_1 are
        the L2 and L1 norms, respectively. This is an abstract class,
        actual implementations need to fill in each method.
    """

    def __init__(self):
        pass

    def initial_guess(self):
        raise Exception('initial_guess method not implemented!')

    def residual(self, x):
        raise Exception('residual method not implemented!')

    def usefulness(self, x, indices):
        """ Given a value x and a set of indices compute the "usefulness" of
            each index, i.e. which should be added to the active set (the "in crowd").
            Return a list of the same size as "indices" where each element
            has the corresponding usefulness.
        """
        raise Exception('usefulness method not implemented!')

    def submatrix(self, indices):
        """ Retrieves a submatrix of A generated by
            the columns of A indexed by "indices".
        """
        raise Exception('submatrix method not implemented!')

    def target(self):
        """ Retrieve the target output vector """
        raise Exception('target method not implemented!')



class FullInCrowdModel(InCrowdModel):
    """ This implementation uses an explicit form for the matrix A and the response vector y. """

    def __init__(self, A, y):
        """ A: a numpy.array object with the # of cols = # of params
            y: a numpy.array object with # of elements = # of data points = # of rows of A
        """
        InCrowdModel.__init__(self)

        self.A = A
        self.y = y
        self.num_params = self.A.shape[1]

    def initial_guess(self):
        return np.zeros(self.num_params)

    def residual(self, x):
        yhat = np.dot(self.A, x)
        #print('yhat=',yhat)
        r = self.y - yhat
        #print('r=',r)
        return r


    def usefulness(self, x, indices):
        """ Given a value x and a set of indices compute the "usefulness" of
            each index, i.e. which should be added to the active set (the "in crowd").
            Return a list of the same size as "indices" where each element
            has the corresponding usefulness.
        """
        r = self.residual(x)
        u = np.dot(r, self.A[:, indices])
        return u

    def submatrix(self, indices):
        """ Retrieves a submatrix of A generated by
            the columns of A indexed by "indices".
        """
        return self.A[:, indices]

    def target(self):
        return self.y


class ConvolutionalInCrowdModel(InCrowdModel):

    def __init__(self, input, target, ndelays=None, bias=0.0, group_index=None, lags=None):
        InCrowdModel.__init__(self)
        if lags is not None:
            ndelays = len(lags)
        self.ndelays = ndelays
        if lags is None:
            self.time_lags = np.arange(0, self.ndelays, 1, dtype='int')
        else:
            self.time_lags = lags
        self.input = input
        self.target_val = target
        self.num_channels = self.input.shape[1]
        self.bias = bias
        self.group_index = group_index

    def initial_guess(self):
        filter = np.zeros([self.num_channels, self.ndelays])
        return filter.reshape(self.num_channels*self.ndelays)

    def get_filter(self, x):
        """ Returns a reshaped filter based on the given vector of parameter values """
        return x.reshape(self.num_channels, self.ndelays)

    def forward(self, x):
        filter = self.get_filter(x)
        if self.group_index is None:
            yhat = fast_conv(self.input, filter, self.time_lags, bias=self.bias)
        else:
            yhat = fast_conv_grouped(self.input, filter, self.time_lags, group_index=self.group_index, bias=self.bias)

        return yhat

    def residual(self, x):
        yhat = self.forward(x)
        r = self.target_val - yhat
        return r

    def usefulness(self, x, indices):
        r = self.residual(x)
        u = np.zeros(len(indices))
        for k,i in enumerate(indices):
            a = self.submatrix([i])
            u[k] = np.dot(r, a)

        return u

    def submatrix(self, indices, include_bias=False):
        A = np.zeros([self.input.shape[0], len(indices)+include_bias])
        if include_bias:
            A[:, -1] = 1.0 # the bias term
        d = len(self.time_lags)
        m = self.num_channels
        all_indices = np.arange(d*m)

        #compute the channel corresponding to each parameter in the reshaped (flattened) filter
        channel_indices = np.floor(all_indices / float(d)).astype('int')

        #compute the lag index corresponding to each parameter in the reshaped (flattened) filter
        lag_indices = all_indices % d
        #print('lag_indices=',lag_indices)

        for k,i in enumerate(indices):
            #get lag and channel corresponding to this index
            lag_index = lag_indices[i]
            #print('k=%d, i=%d, lag_index=%d' % (k, i, lag_index))
            lag = self.time_lags[lag_index]
            channel_to_get = channel_indices[i]

            if lag == 0:
                A[:, k] = self.input[:, channel_to_get]
            else:
                #shift time series for this channel up or down depending on lag
                if lag > 0:
                    A[lag:, k] = self.input[:-lag, channel_to_get]
                else:
                    A[:lag, k] = self.input[-lag:, channel_to_get] #note that lag is negative
        return A

    def submatrix_causal(self, indices):
        A = np.zeros([self.input.shape[0], len(indices)])
        for k,i in enumerate(indices):
            #get channel are we're looking for
            channel_to_get = i % self.num_channels
            #get shift, the number of zeros that precede that channel
            downshift = np.floor(((self.ndelays*self.num_channels) - i - 0.001) / self.num_channels)
            if downshift == 0:
                A[downshift:, k] = self.input[:, channel_to_get]
            else:
                A[downshift:, k] = self.input[:-downshift, channel_to_get]
        return A

    def target(self):
        return self.target_val


def fast_conv_grouped(input, filter, time_lags, group_index, bias=0.0):
    """ Expects input to be a TxM matrix, where M = # of channels, T = # of time points """

    resp = np.zeros(input.shape[0])

    ugroups = np.unique(group_index)
    for grp in ugroups:

        gindex = (group_index == grp).nonzero()[0]
        ginput = input[gindex, :]
        resp[gindex] = fast_conv(ginput, filter, time_lags, bias=bias)

    return resp


def fast_conv(input, filter, time_lags, bias=0.0):
    """ Expects input to be a TxM matrix, where M = # of channels, T = # of time points """

    input_T = np.matrix(input)
    nsamps = input_T.shape[0]
    #print('input_T.shape=',input_T.shape)
    #print('time_lags.shape=',time_lags.shape)
    #print('filter.shape=',filter.shape)

    a = np.zeros( [nsamps, 1] )
    for k,ti in enumerate(time_lags):
        #print('\tti=%d' % ti)
        #print('\tk=%d, filter[:, k].shape=' % k,filter[:, k].shape)
        at = input_T * filter[:, k].reshape(filter.shape[0], 1)
        if ti >= 0:
            if ti > 0:
                at = at[:-ti]
            #print('\tat.shape=',at.shape)
            a[ti:] += at
        else:
            offset = ti % nsamps
            a[:offset] += at[-ti:]

    return a.squeeze() + bias



"""
    Implementation of multi-variate EMD according to:
    N Rehman and DP Mandic "Multivariate empirical mode decomposition"
    Proc. R. Soc. A (2010) 466, 1291-1302 doi:10.1098/rspa.2009.0502



"""
from __future__ import division, print_function

import copy
import numpy as np
from scipy.interpolate import splrep,splev
from scipy.stats import pearsonr
import time
from soundsig.quasirand import quasirand
from soundsig.signal import find_extrema


def create_mirrored_spline(mini, maxi, s):
    """
        To reduce end effects, we need to extend the signal on both sides and reflect the first and last extrema
        so that interpolation works better at the edges
    """

    #get the values of s at the minima and maxima
    s_min = list(s[mini])
    s_max = list(s[maxi])

    #reflect the extrema on the left side
    Tl = maxi[0]  # index of left-hand (first) maximum
    tl = mini[0]  # index of left-hand (first) minimum

    maxi.insert(0, -tl)
    s_max.insert(0, s_max[0])
    mini.insert(0, -Tl)
    s_min.insert(0, s_min[0])

    #reflect the extrema on the right side
    T = len(s)
    Tr = maxi[-1]  # index of right hand (last) maximum
    tr = mini[-1]  # index of right hand (last) minimum

    maxi.append((T-tr) + T)
    s_max.append(s_max[-1])
    mini.append((T-Tr) + T)
    s_min.append(s_min[-1])

    if len(maxi) < 4 or len(mini) < 4:
        return None, None

    #interpolate the upper and lower envelopes
    upper_env_spline = splrep(maxi, s_max, k=3)
    lower_env_spline = splrep(mini, s_min, k=3)

    return lower_env_spline,upper_env_spline


def compute_mean_envelope(s, nsamps=1000):
    """ Use random sampling to compute the mean envelope of a multi-dimensional signal.

    Args:
        s (np.ndarray): an NxT matrix describing a multi-variate signal. N is the number of channels, T is the number of time points.
        nsamps (int): the number of N dimensional projections to use in computing the multi-variate envelope.

    Returns:
        env (np.ndarray): an NxT matrix giving the multi-dimensional envelope of s.
    """

    N,T = s.shape

    #pre-allocate the mean envelope matrix
    mean_env = np.zeros([N, T])

    #generate quasi-random points on an N-dimensional sphere
    #stime = time.time()
    R = quasirand(N, nsamps, spherical=True)
    #etime = time.time() - stime
    #print 'Elapsed time for quasirand: %d seconds' % int(etime)

    stime = time.time()
    for k in range(nsamps):
        #istime = time.time()
        r = R[:, k].squeeze()

        #print 'k=%d, s.shape=%s, r.shape=%s' % (k, str(s.shape), str(r.shape))

        #project s onto a scalar time series using random vector
        #dtime = time.time()
        p = np.dot(s.T, r)
        #detime = time.time() - dtime
        #print '\t[%d] Time to do dot %0.6f s' % (k, detime)

        #print 'p.shape=',p.shape

        #identify minima and maxima of projection
        #eetime = time.time()
        mini_p,maxi_p = find_extrema(p)
        #eeetime = time.time() - eetime
        #print '\t[%d] time to find extrema: %0.6fs' % (k, eeetime)

        #for each signal dimension, fit maxima with cubic spline to produce envelope

        t = np.arange(T)
        for n in range(N):
            #sptime = time.time()
            mini = copy.copy(mini_p)
            maxi = copy.copy(maxi_p)
            if len(mini) < 4 or len(maxi) < 4:
                return None

            #extrapolate edges using mirroring
            lower_env_spline, upper_env_spline = create_mirrored_spline(mini, maxi, s[n, :].squeeze())
            if lower_env_spline is None or upper_env_spline is None:
                return None

            #evaluate upper and lower envelopes
            upper_env = splev(t, upper_env_spline)
            lower_env = splev(t, lower_env_spline)

            #compute the envelope for this projected dimension
            env = (upper_env + lower_env) / 2.0

            #update the mean envelope for this dimension in an online way
            delta = env - mean_env[n, :]
            mean_env[n, :] += delta / (k+1.0)
            #esptime = time.time() - sptime
            #print '\t[%d] time for spline iteration on dimension %d: %0.6fs' % (k, n, esptime)
        #ietime = time.time() - istime
        #print '\t[%d] took %0.6f seconds' % (k, ietime)

    etime = time.time() - stime
    #print '%d samples took %0.6f seconds' % (nsamps, etime)

    return mean_env


def sift(s, nsamps=100, resolution=50.0, max_iterations=30, verbose=False):
    """Do a single iteration of multi-variate empirical mode decomposition (MEMD) on the multi-dimensional signal s, obtaining a multi-variate IMF.

    Args:
        s (np.ndarray): an NxT matrix describing a multi-variate signal. N is the number of channels, T is the number of time points.
        nsamps (int): the number of N dimensional projections to use in computing the multi-variate envelope.
        resolution (float): the maximum log10 ratio of the initial signal energy to average envelope energy, used as stopping criteria (Rato et. al 2008, sec 3.2.3)
        max_iterations (int): the maximum number of iterations before quitting

    Returns:
        imf (np.ndarray): an NxT matrix giving the multi-dimensional IMF for this sift.

    """

    converged = False

    # inintialize the residual signal to the signal
    N,T = s.shape
    r = copy.copy(s)

    initial_energy = (s**2).mean()
    avg_envelope_energy = 0.0
    iteration = 1
    while not converged:
        # compute the mean envelope
        env = compute_mean_envelope(r, nsamps=nsamps)
        if env is None:
            converged = True
            break

        #update the average envelope energy
        env_energy = (env**2).mean()
        avg_envelope_energy += (env_energy - avg_envelope_energy) / iteration

        # compute the fraction of the mean_envelope that will be subtracted from the residual, meant to minimize the
        # energy of the signal (Rato et. al 2008, sec 3.2.3)
        alpha = np.zeros([N])
        for k in range(N):
            a,p = pearsonr(r[k, :], env[k, :])
            if np.isnan(a):
                print('[iter %d] alpha=NaN for dimension %d!' % (iteration, k))
                a = 1e-2
            alpha[k] = max(a, 1e-2)

        # subtract the mean envelope from the residual
        final_alpha = alpha.mean()
        r -= alpha.mean()*env

        # test the residual for convergence to an IMF using stoppage criteria

        #compute the "resolution factor", the ratio of initial energy to average envelope energy. the higher the
        #"resolution", the lower the average envelope energy, meaning that the IMF is converging.
        resolution_factor = np.log10(initial_energy / avg_envelope_energy)

        if verbose:
            print('sift iter %d: initial_energy=%0.3f, env_energy=%0.3f, avg_envelope_energy=%0.3f, final_alpha=%0.2f, resolution_factor=%0.2f' %
                  (iteration, initial_energy, env_energy, avg_envelope_energy, final_alpha, resolution_factor))

        if resolution_factor > resolution:
            converged = True
        if iteration >= max_iterations:
            converged = True

        iteration += 1

    # after repeatedly subtracting off the mean envelope, we are left with the intrinsic mode function (IMF), which
    # basically contains the higher frequency components of the original signal, with the lower frequency components
    # subtracted off in a way that deals well with nonstationarity
    return r


def memd(s, nimfs, nsamps=100, resolution=1.0, max_iterations=30, num_noise_channels=0, verbose=False):

    imfs = list()
    N,T = s.shape
    if verbose:
        print('Starting MEMD: N=%d, T=%d, nimfs=%d, nsamps=%d, resolution=%0.2f, max_iterations=%d, num_noise_channels=%d' %
                        (N, T, nimfs, nsamps, resolution, max_iterations, num_noise_channels))
    r = copy.copy(s)

    if num_noise_channels > 0:
        #add gaussian noise channels for Noise-assisted MEMD (2011 Rehnert and Mandic)
        noise = np.random.randn(num_noise_channels, T)
        r = np.vstack([r, noise])

    for n in range(nimfs):
        imf = sift(r, nsamps=nsamps, resolution=resolution, max_iterations=max_iterations, verbose=verbose)
        imfs.append(imf)
        r -= imf
    imfs = np.array(imfs)

    return imfs[:, :N, :]



from __future__ import division

import numpy as np


class DiscreteProcessEstimator(object):
    """ This class estimates the emperical distribution of a discrete random variable over time. It forgets the
        distribution with a user-specified time constant.
    """

    def __init__(self, num_states, sample_rate, time_constant):
        self.num_states = num_states
        self.state = np.zeros([num_states])
        self.sample_rate = sample_rate
        self.p = np.zeros([num_states])
        self.tau = time_constant
        self.dt = 1. / self.sample_rate
        self.eps = 1e-1

    def update(self, new_observation):
        """ Update the state of the system with a new observation.

        :param new_observation: An integer from 1 to self.num_states, indicated the category (level) of the
               variable (factor). A zero indicates that there is no observation for this time period.

        :return:
        """

        # convert the observation to one-of-k encoding
        v = np.zeros([self.num_states])
        if new_observation > 0:
            v[new_observation-1] = 1.
        self.state = self.state * (1. - (self.dt / self.tau)) + v
        self.p = (self.state + self.eps) / np.sum(self.state + self.eps)


class Discretizer(object):
    """ This class discretizes a scalar variable.
    """

    def __init__(self, xvals, num_levels=3):
        """ Initialize the class.

        :param xvals: An np.array of observations of the scalar variable.
        """
        self.num_levels = num_levels

        # choose the percentiles that separate levels
        dq = 1. / num_levels
        self.percentiles = np.zeros([num_levels-1], dtype='int')
        self.percentile_vals = np.zeros([num_levels-1])
        for k in range(num_levels-1):
            self.percentiles[k] = int((k+1)*dq*100)
            self.percentile_vals[k] = np.percentile(xvals, self.percentiles[k])

    def discretize(self, xval):
        """ Turn the scalar variable xval into the discrete category to which it belongs.

        :param xval:
        :return:
        """
        i = np.where(xval > self.percentile_vals)[0]
        if len(i) == 0:
            return 1
        else:
            return max(i) + 2



from __future__ import division

import operator

import numpy as np

import matplotlib.pyplot as plt
from soundsig.plots import make_phase_image


class ComplexPCA(object):

    def __init__(self, n_components=None):
        self.n_components = n_components

    def fit(self, Z, zscore=True):
        """ Does principal components analysis on a complex feature matrix.

        :param Z: A complex-valued matrix of shape (num_samples, num_features)
        :param n_components: The number of principal components to compute.
        """

        num_samples, num_features = Z.shape

        if zscore:
            #center the data
            Z.real -= Z.real.mean(axis=0)
            Z.imag -= Z.imag.mean(axis=0)

            #rescale the data so real and imaginary parts have unit std
            Z.real /= Z.real.std(axis=0, ddof=1)
            Z.imag /= Z.imag.std(axis=0, ddof=1)

        self.Z = Z

        #compute covariance
        self.C = np.dot(np.conj(Z.T), Z) / num_samples

        #compute pseudo-covariance
        self.Cp = np.dot(Z.T, Z) / num_samples

        #compute eigenspectrum of covariance
        self.v,self.U = np.linalg.eig(self.C)
        self.v,self.U = sort_eigenspectrum(self.v,self.U)

        #compute eigenspectrum of pseudo-covariance
        self.vp,self.Up = np.linalg.eig(self.Cp)
        self.vp,self.Up = sort_eigenspectrum(self.vp,self.Up)

    def plot(self):
                #make plots of real and imaginary parts, phase and amplitudes
        plt.figure()
        gs = plt.GridSpec(2, 2)

        ax = plt.subplot(gs[0, 0])
        plt.hist(self.Z.real.ravel(), bins=50, color='r')
        plt.title('Distribution of Real Parts')
        plt.axis('tight')

        ax = plt.subplot(gs[0, 1])
        plt.hist(self.Z.imag.ravel(), bins=50, color='m')
        plt.title('Distribution of Imaginary Parts')
        plt.axis('tight')

        ax = plt.subplot(gs[1, 0])
        plt.hist(np.abs(self.Z).ravel(), bins=50, color='k')
        plt.title('Distribution of Amplitudes')
        plt.axis('tight')

        ax = plt.subplot(gs[1, 1])
        plt.hist((180.0/np.pi)*np.angle(self.Z).ravel(), bins=50, color='b')
        plt.axis('tight')
        plt.title('Distribution of Phases')

        #plot the covariance matrices
        img = make_phase_image(np.abs(self.C), np.angle(self.C))
        imgp = make_phase_image(np.abs(self.Cp), np.angle(self.Cp))

        plt.figure()
        gs = plt.GridSpec(2, 2)

        ax = plt.subplot(gs[:2, 0])
        ax.set_axis_bgcolor('black')
        plt.imshow(img, interpolation='nearest', aspect='auto')
        plt.title('Covariance')

        ax = plt.subplot(gs[:2, 1])
        ax.set_axis_bgcolor('black')
        plt.imshow(imgp, interpolation='nearest', aspect='auto')
        plt.title('Pseudo-Covariance')

        #plot the eigenvalues
        plt.figure()

        gs = plt.GridSpec(3, 2)

        ax = plt.subplot(gs[0, 0])
        plt.plot(np.abs(self.v), 'go')
        ax.set_yscale('log')
        plt.title('Covariance Eigenvalues')
        plt.axis('tight')

        ax = plt.subplot(gs[0, 1])
        plt.plot(np.abs(self.vp), 'go')
        ax.set_yscale('log')
        plt.title('Pseudo-Covariance Eigenvalues')
        plt.axis('tight')

        #plot the principal components
        ax = plt.subplot(gs[1:, 0])
        ax.set_axis_bgcolor('black')
        eimg = make_phase_image(np.abs(self.U), np.angle(self.U))
        plt.imshow(eimg, interpolation='nearest', aspect='auto')

        ax = plt.subplot(gs[1:, 1])
        ax.set_axis_bgcolor('black')
        eimgp = make_phase_image(np.abs(self.Up), np.angle(self.Up))
        plt.imshow(eimgp, interpolation='nearest', aspect='auto')


    def transform(self, Z):
        return np.dot(Z, self.U[:, :self.n_components])

    def inverse_transform(self, Z):
        return np.dot(Z, self.U[:, :self.n_components].T)


def sort_eigenspectrum(eigenvalues, eigenvectors):

    elist = [(k,np.abs(v)) for k,v in enumerate(eigenvalues)]
    elist.sort(key=operator.itemgetter(1), reverse=True)
    index = [x[0] for x in elist]
    eigenvalues = eigenvalues[index]
    eigenvectors = eigenvectors[:, index]

    return eigenvalues,eigenvectors






from __future__ import division, print_function

import copy
import fnmatch
import os
import subprocess
import wave
import struct
import hashlib
import h5py
from copy import deepcopy
from math import ceil
from numpy.fft import fftshift, ifftshift

import numpy as np
from scipy.io.wavfile import read as read_wavfile
from scipy.fftpack import fft, ifft, fftfreq, fft2, ifft2, dct
from scipy.signal import resample, firwin, filtfilt
from scipy.linalg import inv, toeplitz
from scipy.optimize import leastsq


import matplotlib.pyplot as plt
import matplotlib.cm as cmap
import matplotlib.colors as pltcolors
import matplotlib.mlab as mlab

import colorsys
from soundsig.signal import lowpass_filter, gaussian_window, correlation_function
from soundsig.timefreq import gaussian_stft
from soundsig.detect_peaks import detect_peaks


class WavFile():
    """ Class for representing a sound and writing it to a .wav file """

    def __init__(self, file_name=None, log_spectrogram=True):

        self.log_spectrogram = log_spectrogram
        if file_name is None:
            self.sample_depth = 2  # in bytes
            self.sample_rate = 44100.0  # in Hz
            self.data = None
            self.num_channels = 1
        else:
            wr = wave.open(file_name, 'r')
            self.num_channels = wr.getnchannels()
            self.sample_depth = wr.getsampwidth()
            wr.close()

            self.sample_rate, data = read_wavfile(file_name)
            # If stereo make mono
            if self.num_channels == 1:
                self.data = data
            else:
                self.data = data.mean(axis=1)

        self.analyzed = False

    def to_wav(self, output_file, normalize=False, max_amplitude=32767.0):
        wf = wave.open(output_file, 'w')

        wf.setparams( (self.num_channels, self.sample_depth, self.sample_rate, len(self.data), 'NONE', 'not compressed') )
        # normalize the sample
        if normalize:
            nsound = ((self.data / np.abs(self.data).max())*max_amplitude).astype('int')
        else:
            nsound = self.data
        #print 'nsound.min=%d, max=%d' % (nsound.min(), nsound.max())
        hex_sound = [struct.pack('h', x) for x in nsound]
        wf.writeframes(b''.join(hex_sound))
        wf.close()

    def analyze(self, min_freq=0, max_freq=None, spec_sample_rate=1000.0, freq_spacing=125.0, envelope_cutoff_freq=200.0, noise_level_db=80, rectify=True, cmplx=False):
        if self.analyzed:
            return

        self.data_t = np.arange(0.0, len(self.data), 1.0) / self.sample_rate

        #compute the temporal envelope
        self.envelope = temporal_envelope(self.data, self.sample_rate, envelope_cutoff_freq)

        #compute log power spectrum
        fftx = fft(self.data)
        ps_f = fftfreq(len(self.data), d=(1.0 / self.sample_rate))
        if max_freq == None:
            findx = (ps_f > min_freq) & (ps_f < np.inf)
        else:
            findx = (ps_f > min_freq) & (ps_f < max_freq)
        self.power_spectrum = np.log10(np.abs(fftx[findx]))
        self.power_spectrum_f = ps_f[findx]

        #estimate fundamental frequency from log power spectrum in the simplest way possible
        ps = np.abs(fftx[findx])
        peak_index = ps.argmax()
        try:
            self.fundamental_freq = self.power_spectrum_f[peak_index]
        except IndexError:
            print('Could not identify fundamental frequency!')
            self.fundamental_freq = 0.0

        #compute log spectrogram
        t,f,spec,spec_rms = spectrogram(self.data, self.sample_rate, spec_sample_rate=spec_sample_rate,
                                        freq_spacing=freq_spacing, min_freq=min_freq, max_freq=max_freq)
        self.spectrogram_t = t
        self.spectrogram_f = f
        self.spectrogram = spec
        self.spectrogram_rms = spec_rms
        self.analyzed = True

    def reanalyze(self, min_freq=0, max_freq=None, spec_sample_rate=1000.0, freq_spacing=25.0, envelope_cutoff_freq=200.0, noise_level_db=80, rectify=True, cmplx=False):
        self.analyzed = False
        return self.analyze(min_freq, max_freq, spec_sample_rate, freq_spacing, envelope_cutoff_freq, noise_level_db, rectify, cmplx)

    def plot(self, fig=None, show_envelope=True, min_freq=0.0, max_freq=10000.0, colormap=cmap.gist_yarg, noise_level_db=80,
             start_time=0, end_time=np.inf):

        self.analyze(min_freq=min_freq, max_freq=max_freq, noise_level_db=noise_level_db)

        if show_envelope:
            spw_size = 15
            spec_size = 35
        else:
            spw_size = 25
            spec_size = 75

        raw_ti = (self.data_t > start_time) & (self.data_t < end_time)

        if fig is None:
            fig = plt.figure()
        gs = plt.GridSpec(100, 1)
        ax = fig.add_subplot(gs[:spw_size])
        plt.plot(self.data_t[raw_ti], self.data[raw_ti], 'k-')
        plt.axis('tight')
        plt.ylabel('Sound Pressure')

        s = (spw_size+5)
        e = s + spec_size
        ax = fig.add_subplot(gs[s:e])
        spec_ti = (self.spectrogram_t > start_time) & (self.spectrogram_t < end_time)
        plot_spectrogram(self.spectrogram_t[spec_ti], self.spectrogram_f, self.spectrogram[:, spec_ti], ax=ax, ticks=True, colormap=colormap, colorbar=False)

        if show_envelope:
            ax = fig.add_subplot(gs[(e+5):95])
            plt.plot(self.spectrogram_t, self.spectrogram_rms, 'g-')
            plt.xlabel('Time (s)')
            plt.ylabel('Envelope')
            plt.axis('tight')
            
class BioSound(object):
    """ Class for representing a communication sound using multiple feature spaces"""

    def __init__(self, soundWave=np.array(0.0), fs=np.array(0.0), emitter='Unknown', calltype = 'U' ):
        # Note that all the fields are numpy arrays for saving to h5 files.

        self.sound = soundWave  # sound pressure waveform 
        self.hashid = np.string_(hashlib.md5(np.array_str(soundWave).encode('utf-8')).hexdigest())
        self.samprate = float(fs) if isinstance(fs,int) else fs      # sampling rate
        self.emitter = np.string_(emitter)  # string for id of emitter
        self.type = np.string_(calltype)    # string for call type
        self.spectro = np.asarray([])    # Log spectrogram
        self.to = np.asarray([])         # Time scale for spectrogram
        self.fo = np.asarray([])         # Frequency scale for spectrogram
        self.mps = np.asarray([])        # Modulation Power Spectrum
        self.wf = np.asarray([])         # Spectral modulations
        self.wt = np.asarray([])         # Temporal modulations
        self.f0 = np.asarray([])         # time varying fundamental
        self.f0_2 = np.asarray([])       # time varying fundamental of second voice
        self.F1 = np.asarray([])         # time varying formant 1
        self.F2 = np.asarray([])         # time varying formant 2
        self.F3 = np.asarray([])         # time varying formant 3
        self.fund = np.asarray([])       # Average fundamental
        self.sal = np.asarray([])        # Average saliency
        self.fund2 = np.asarray([])      # Average fundamental of 2nd peak
        self.voice2percent = np.asarray([]) # Average percent of presence of second peak
        self.maxfund = np.asarray([])
        self.minfund = np.asarray([])
        self.cvfund = np.asarray([])
        self.meanspect = np.asarray([])
        self.stdspect = np.asarray([])
        self.skewspect = np.asarray([])
        self.kurtosisspect = np.asarray([])
        self.entropyspect = np.asarray([])
        self.q1 = np.asarray([])
        self.q2 = np.asarray([])
        self.q3 = np.asarray([])
        self.meantime = np.asarray([])
        self.stdtime = np.asarray([])
        self.skewtime = np.asarray([])
        self.kurtosistime = np.asarray([])
        self.entropytime = np.asarray([])
        self.fpsd = np.asarray([])
        self.psd = np.asarray([])
        self.tAmp = np.asarray([])
        self.amp = np.asarray([])
        self.rms = np.asarray([])
        self.maxAmp = np.asarray([])
        
    def saveh5(self, fileName=None):
   # Save as an h5 file. Uses the hashid if fileName is not given
   # Not using attributes
   
        if fileName is None:
            fileName = '%s.h5' % self.hashid
            
        fid = h5py.File(fileName,'w')
        selfDict = vars(self)
        for varnames in selfDict:
            fid.create_dataset(varnames, data=selfDict[varnames])
            
        fid.close()
        
    def readh5(self, fileName):
        
        fid = h5py.File(fileName, 'r')
        for varnames in fid.keys():
            setattr(self, varnames, np.array(fid[varnames]).squeeze())
        
        fid.close()
        
    def spectrum(self, f_high=10000):
    # Calculates power spectrum and features from power spectrum

    # Need to add argument for window size
    # f_high is the upper bound of the frequency for saving power spectrum
    # nwindow = (1000.0*np.size(soundIn)/samprate)/window_len
    # 
        Pxx, Freqs = mlab.psd(self.sound, Fs=self.samprate, NFFT=1024, noverlap=512)
    
        # Find quartile power
        cum_power = np.cumsum(Pxx)
        tot_power = np.sum(Pxx)
        quartile_freq = np.zeros(3, dtype = 'int')
        quartile_values = [0.25, 0.5, 0.75]
        nfreqs = np.size(cum_power)
        iq = 0
        for ifreq in range(nfreqs):
            if (cum_power[ifreq] > quartile_values[iq]*tot_power):
                quartile_freq[iq] = ifreq
                iq = iq+1
                if (iq > 2):
                    break
                 
        # Find skewness, kurtosis and entropy for power spectrum below f_high
        ind_fmax = np.where(Freqs > f_high)[0][0]
    
        # Description of spectral shape
        spectdata = Pxx[0:ind_fmax]
        freqdata = Freqs[0:ind_fmax]
        spectdata = spectdata/np.sum(spectdata)
        meanspect = np.sum(freqdata*spectdata)
        stdspect = np.sqrt(np.sum(spectdata*((freqdata-meanspect)**2)))
        skewspect = np.sum(spectdata*(freqdata-meanspect)**3)
        skewspect = skewspect/(stdspect**3)
        kurtosisspect = np.sum(spectdata*(freqdata-meanspect)**4)
        kurtosisspect = kurtosisspect/(stdspect**4)
        entropyspect = -np.sum(spectdata*np.log2(spectdata))/np.log2(ind_fmax)
 
        # Storing the values       
        self.meanspect = meanspect
        self.stdspect = stdspect
        self.skewspect = skewspect
        self.kurtosisspect = kurtosisspect
        self.entropyspect = entropyspect
        self.q1 = Freqs[quartile_freq[0]]
        self.q2 = Freqs[quartile_freq[1]]
        self.q3 = Freqs[quartile_freq[2]]
        self.fpsd = freqdata
        self.psd = spectdata
        
    def spectroCalc(self, spec_sample_rate=1000, freq_spacing = 50, min_freq=0, max_freq=10000):
        # Calculates the spectrogram in dB
        t,f,spec,spec_rms = spectrogram(self.sound, self.samprate, spec_sample_rate=spec_sample_rate,
                                        freq_spacing=freq_spacing, min_freq=min_freq, max_freq=max_freq,
                                        cmplx=True)
        self.to = t
        self.fo = f
        self.spectro = 20*np.log10(np.abs(spec))

    def mpsCalc(self, window=None, Norm=True): 
        
        if self.spectro.size == 0:
            self.spectroCalc()
            
        wf, wt, mps_powAvg = mps(self.spectro, self.fo, self.to, window=window, Norm=Norm)
        self.mps = mps_powAvg        # Modulation Power Spectrum
        self.wf = wf         # Spectral modulations
        self.wt = wt
        
        
    def ampenv(self, cutoff_freq = 20, amp_sample_rate = 1000):
    # Calculates the amplitude enveloppe and related parameters
    
        (amp, tdata)  = temporal_envelope(self.sound, self.samprate, cutoff_freq=cutoff_freq, resample_rate=amp_sample_rate)
        
        # Here are the parameters
        ampdata = amp/np.sum(amp)
        meantime = np.sum(tdata*ampdata)
        stdtime = np.sqrt(np.sum(ampdata*((tdata-meantime)**2)))
        skewtime = np.sum(ampdata*(tdata-meantime)**3)
        skewtime = skewtime/(stdtime**3)
        kurtosistime = np.sum(ampdata*(tdata-meantime)**4)
        kurtosistime = kurtosistime/(stdtime**4)
        indpos = np.where(ampdata>0)[0]
        entropytime = -np.sum(ampdata[indpos]*np.log2(ampdata[indpos]))/np.log2(np.size(indpos))
        
        self.meantime = meantime   
        self.stdtime = stdtime
        self.skewtime = skewtime
        self.kurtosistime = kurtosistime
        self.entropytime = entropytime
        self.tAmp = tdata
        self.amp = amp
        self.maxAmp = max(amp)
        
    def fundest(self, maxFund = 1500, minFund = 300, lowFc = 200, highFc = 6000, minSaliency = 0.5, debugFig = 0, minFormantFreq = 500, maxFormantBW = 500, method='Stack'):
    # Calculate the fundamental, the formants and parameters related to these
    
        sal, fund, fund2, form1, form2, form3, lenfund = fundEstimator(self.sound, self.samprate, self.to, debugFig = debugFig, maxFund = maxFund, minFund = minFund, lowFc = lowFc, highFc = highFc, minSaliency = minSaliency, minFormantFreq = minFormantFreq, maxFormantBW = maxFormantBW, method = method)
        goodFund = fund[~np.isnan(fund)]
        goodSal = sal[~np.isnan(sal)]
        goodFund2 = fund2[~np.isnan(fund2)]
        if np.size(goodFund) > 0 :
            meanfund = np.mean(goodFund)
        else:
            meanfund = np.asarray([])
        meansal = np.mean(goodSal)
        if np.size(goodFund2)> 0:
            meanfund2 = np.mean(goodFund2)
        else:
            meanfund2 = np.asarray([])
    
        if np.size(goodFund) == 0 or np.size(goodFund2) == 0:
            fund2prop = 0.0
        else:
            fund2prop = np.float(np.size(goodFund2))/np.float(np.size(goodFund))
            
        self.f0 = fund         # time varying fundamental
        self.f0_2 = fund2        # time varying fundamental of second voice
        self.F1 = form1         # time varying formant 1
        self.F2 = form2         # time varying formant 2
        self.F3 = form3         # time varying formant 3
        self.fund = meanfund       # Average fundamental
        self.sal = meansal        # Average saliency
        self.fund2 = meanfund2      # Average fundamental of 2nd peak
        self.voice2percent = fund2prop*100 # Average percent of presence of second peak
        if np.size(goodFund) > 0 :
            self.maxfund = np.max(goodFund)
            self.minfund = np.min(goodFund)
            self.cvfund = np.std(goodFund)/meanfund
            
    def play(self):
    # Plays the sound
        play_sound_array(self.sound*(2**15), self.samprate)
            
    def plot(self, DBNOISE=50, f_low=250, f_high=10000):
    # Plots a biosound in figures 1, 2, 3
    
        # Ploting Variables
        soundlen = np.size(self.sound)
        t = np.array(range(soundlen))
        t = t*(1000.0/self.samprate)

        # Plot the oscillogram + spectrogram
        plt.figure(1)
        plt.clf()
        # mngr = plt.get_current_fig_manager()
        # mngr.window.setGeometry(0, 260, 640, 545)
        
        
        # The oscillogram
        plt.axes([0.1, 0.75, 0.85, 0.20])      
        plt.plot(t,self.sound, 'k')
        # plt.xlabel('Time (ms)')
        plt.xlim(0, t[-1])               
        # Plot the amplitude enveloppe  
        if self.tAmp.size != 0 :   
            # rescale amp envelope to max for better display
            plt.plot(self.tAmp*1000.0, self.amp*np.max(self.sound)/np.max(self.amp), 'r', linewidth=2)
      
        # Plot the spectrogram
        plt.axes([0.1, 0.1, 0.85, 0.6])
        spec_colormap()   # defined in sound.py
        cmap = plt.get_cmap('SpectroColorMap')
        
        if self.spectro.size != 0 :
            soundSpect = self.spectro
            if soundSpect.shape[0] == self.to.size:
                soundSpect = np.transpose(soundSpect)
            maxB = soundSpect.max()
            minB = maxB-DBNOISE
            soundSpect[soundSpect < minB] = minB
            minSpect = soundSpect.min()
            plt.imshow(soundSpect, extent = (self.to[0]*1000, self.to[-1]*1000, self.fo[0], self.fo[-1]), aspect='auto', interpolation='nearest', origin='lower', cmap=cmap, vmin=minSpect, vmax=maxB)
        
        plt.ylim(f_low, f_high)
        plt.xlim(0, t[-1])
        plt.ylabel('Frequency (Hz)')
        plt.xlabel('Time (ms)')
                     
    # Plot the fundamental on the same figure
        if self.f0.size != 0 :
            fundplot = self.f0
            diffFund = np.diff(fundplot)
            diffFundInd = np.concatenate(([False], abs(diffFund)>1000))
            fundplot[diffFundInd] = float('nan')
            plt.plot(self.to*1000.0, self.f0, 'k', linewidth=3)
            plt.plot(self.to*1000.0, self.f0_2, 'm', linewidth=3)
            plt.plot(self.to*1000.0, self.F1, 'r--', linewidth=3)
            plt.plot(self.to*1000.0, self.F2, 'w--', linewidth=3)
            plt.plot(self.to*1000.0, self.F3, 'b--', linewidth=3)
        plt.show()
           
    # Plot Power Spectrum
        plt.figure(2)
        plt.clf()
        # mngr = plt.get_current_fig_manager()
        # mngr.window.setGeometry(650, 260, 640, 545)
        if self.psd.size != 0 :
            plt.plot(self.fpsd, self.psd, 'k-') 
            plt.xlabel('Frequency Hz')
            plt.ylabel('Power Linear')
        
            xl, xh, yl, yh = plt.axis()
            xl = f_low
            xh = f_high
            plt.axis((xl, xh, yl, yh))
            
            if self.q1.size != 0:
                plt.plot([self.q1, self.q1], [yl, yh], 'k--')
                plt.plot([self.q2, self.q2], [yl, yh], 'k--')
                plt.plot([self.q3, self.q3], [yl, yh], 'k--')
                
            if self.F1.size != 0:        
                F1Mean = self.F1[~np.isnan(self.F1)].mean()
                F2Mean = self.F2[~np.isnan(self.F2)].mean()
                F3Mean = self.F3[~np.isnan(self.F3)].mean()
                plt.plot([F1Mean, F1Mean], [yl, yh], 'r--', linewidth=2.0)
                plt.plot([F2Mean, F2Mean], [yl, yh], 'c--', linewidth=2.0)
                plt.plot([F3Mean, F3Mean], [yl, yh], 'b--', linewidth=2.0)
                
            plt.show()
  
    # Table of results
        plt.figure(3)
        plt.clf()
        # mngr = plt.get_current_fig_manager()
        # mngr.window.setGeometry(320, 10, 640, 250)
        textstr = '%s  %s' % (self.emitter, self.type)
        plt.text(0.4, 1.0, textstr)
        if self.fund.size != 0:
            if self.fund2.size != 0:
                textstr = 'Mean Fund = %.2f Hz Mean Saliency = %.2f Mean Fund2 = %.2f PF2 = %.2f%%' % (self.fund, self.sal, self.fund2, self.voice2percent)
            else:
                textstr = 'Mean Fund = %.2f Hz Mean Saliency = %.2f No 2nd Voice Detected' % (self.fund, self.sal)
            plt.text(-0.1, 0.8, textstr)
            
        if self.fund.size != 0:
            textstr = '   Max Fund = %.2f Hz, Min Fund = %.2f Hz, CV = %.2f' % (self.maxfund, self.minfund, self.cvfund) 
            plt.text(-0.1, 0.7, textstr)
        textstr = 'Mean Spect = %.2f Hz, Std Spect= %.2f Hz' % (self.meanspect, self.stdspect)
        plt.text(-0.1, 0.6, textstr)
        textstr = '   Skew = %.2f, Kurtosis = %.2f Entropy=%.2f' % (self.skewspect, self.kurtosisspect, self.entropyspect)
        plt.text(-0.1, 0.5, textstr)
        textstr = '   Q1 F = %.2f Hz, Q2 F= %.2f Hz, Q3 F= %.2f Hz' % (self.q1, self.q2, self.q3 )
        plt.text(-0.1, 0.4, textstr)
        if self.F1.size != 0:
            textstr = '   For1 = %.2f Hz, For2 = %.2f Hz, For3= %.2f Hz' % (F1Mean, F2Mean, F3Mean )
            plt.text(-0.1, 0.3, textstr)
        textstr = 'Mean Time = %.2f s, Std Time= %.2f s' % (self.meantime, self.stdtime)
        plt.text(-0.1, 0.2, textstr)
        textstr = '   Skew = %.2f, Kurtosis = %.2f Entropy=%.2f' % (self.skewtime, self.kurtosistime, self.entropytime)
        plt.text(-0.1, 0.1, textstr)
        if self.rms.size != 0 and self.maxAmp.size != 0 :
            textstr = 'RMS = %.2f, Max Amp = %.2f' % (self.rms, self.maxAmp)
            plt.text(-0.1, 0.0, textstr)
        
        plt.axis('off')        
        plt.show()
        
    # Plot Modulation Power spectrum if it exists
    
        #ex = (spectral_freq.min(), spectral_freq.max(), temporal_freq.min(), temporal_freq.max())
        if self.mps.size != 0 :
            plt.figure(4)
            plt.clf()
            cmap = plt.get_cmap('jet')
            ex = (self.wt.min(), self.wt.max(), self.wf.min()*1e3, self.wf.max()*1e3)
            logMPS = 10.0*np.log10(self.mps)
            maxMPS = logMPS.max()
            minMPS = maxMPS-DBNOISE
            logMPS[logMPS < minMPS] = minMPS
            plt.imshow(logMPS, interpolation='nearest', aspect='auto', origin='lower', cmap=cmap, extent=ex)
            plt.ylabel('Spectral Frequency (Cycles/KHz)')
            plt.xlabel('Temporal Frequency (Hz)')
            plt.colorbar()
            plt.ylim((0,self.wf.max()*1e3))
            plt.title('Modulation Power Spectrum')
            plt.show()
        
        
        plt.pause(1)   # To flush the plots?



def spec_colormap():
# Makes the colormap that we like for spectrograms

    cmap = np.zeros((64,3))
    cmap[0,2] = 1.0

    for ib in range(21):
        cmap[ib+1,0] = (31.0+ib*(12.0/20.0))/60.0
        cmap[ib+1,1] = (ib+1.0)/21.0
        cmap[ib+1,2] = 1.0

    for ig in range(21):
        cmap[ig+ib+1,0] = (21.0-(ig)*(12.0/20.0))/60.0
        cmap[ig+ib+1,1] = 1.0
        cmap[ig+ib+1,2] = 0.5+(ig)*(0.3/20.0)

    for ir in range(21):
        cmap[ir+ig+ib+1,0] = (8.0-(ir)*(7.0/20.0))/60.0
        cmap[ir+ig+ib+1,1] = 0.5 + (ir)*(0.5/20.0)
        cmap[ir+ig+ib+1,2] = 1

    for ic in range(64):
        (cmap[ic,0], cmap[ic,1], cmap[ic,2]) = colorsys.hsv_to_rgb(cmap[ic,0], cmap[ic,1], cmap[ic,2])
    
    spec_cmap = pltcolors.ListedColormap(cmap, name=u'SpectroColorMap', N=64)
    plt.register_cmap(cmap=spec_cmap)

def plot_spectrogram(t, freq, spec, ax=None, ticks=True, fmin=None, fmax=None, colormap=None, colorbar=True, log = True, dBNoise = 50):
    
    if colormap == None:
        spec_colormap()
        colormap = plt.get_cmap('SpectroColorMap')
        
    if ax is None:
        ax = plt.gca()

    if fmin is None:
        fmin = freq.min()
    if fmax is None:
        fmax = freq.max()

    ex = (t.min(), t.max(), freq.min(), freq.max())
    plotSpect = np.abs(spec)
    
    
    if log == True and dBNoise is not None:
        plotSpect = 20*np.log10(plotSpect)
        maxB = plotSpect.max()
        minB = maxB-dBNoise
    else:
        if dBNoise is not None:
            maxB = 20*np.log10(plotSpect.max())
            minB = ((maxB-dBNoise)/20.0)**10
        else:
            maxB = plotSpect.max()
            minB = plotSpect.min()

    plotSpect[plotSpect < minB] = minB
                
    iax = ax.imshow(plotSpect, aspect='auto', interpolation='nearest', origin='lower', extent=ex, cmap=colormap, vmin=minB, vmax=maxB)
    ax.set_ylim(fmin, fmax)
    if not ticks:
        ax.set_xticks([])
        ax.set_yticks([])
    else:
        ax.set_ylabel('Frequency (Hz)')
        ax.set_xlabel('Time (s)')

    if colorbar:
        plt.colorbar(iax)


def play_sound(file_name):
    """ Install sox to get this to work: http://sox.sourceforge.net/ """
    subprocess.call(['play', file_name])

def play_wavfile(filename):

    import pyaudio

    chunk_size = 1024

    wf = wave.open(filename, "r")
    p = pyaudio.PyAudio()
    stream = p.open(format=p.get_format_from_width(wf.getsampwidth()),
        channels=wf.getnchannels(),
        rate=wf.getframerate(),
        output=True)

    data = wf.readframes(chunk_size)

    while data != '':
        stream.write(data)
        data = wf.readframes(chunk_size)

    wf.close()
    stream.stop_stream()
    stream.close()
    p.terminate()

def play_sound_array(data, sample_rate):
    ''' Requires pyaudio package. Can be downloaded here
    http://people.csail.mit.edu/hubert/pyaudio/
    '''

    import pyaudio

    # Only play one channel
    if len(data.shape) > 1:
        data = np.mean(data, axis=np.argmin(data.shape))

    data = data.astype('int16')
    p = pyaudio.PyAudio()
    stream = p.open(format=p.get_format_from_width(2),
        channels=1,
        rate=int(sample_rate),
        output=True)

    stream.write(data.tostring())
    stream.stop_stream()
    stream.close()
    p.terminate()


def spectrogram(s, sample_rate, spec_sample_rate, freq_spacing, min_freq=0, max_freq=None, nstd=6,  cmplx = True):
    """
        Given a sound pressure waveform, s, compute the complex spectrogram. 
        See documentation on gaussian_stft for additional details.

        Returns: 
          t, freq, timefreq, rms
              t: array of time values to use as x axis
              freq: array of frequencies to use as y axis
              timefreq: the spectrogram (a time-frequency represnetaion)
              rms : the time varying average
              
        Arguments:
            REQUIRED:
                s: sound pressssure waveform
                sample_rate: sampling rate for s in Hz
                spec_sample_rate: sampling rate for the output spectrogram in Hz. This variable sets the overlap for the windows in the STFFT.
                freq_spacing: the time-frequency scale for the spectrogram in Hz. This variable determines the width of the gaussian window. 
            
            OPTIONAL            
                complex = False: returns the absolute value
                use min_freq and max_freq to save space
                nstd = number of standard deviations of the gaussian in one window.
    """
     
    # We need units here!!
    increment = 1.0 / spec_sample_rate
    window_length = nstd / (2.0*np.pi*freq_spacing)
    t,freq,timefreq,rms = gaussian_stft(s, sample_rate, window_length, increment, nstd=nstd, min_freq=min_freq, max_freq=max_freq)

    # rms = spec.std(axis=0, ddof=1)
    if cmplx == False:
        timefreq = np.abs(timefreq)
        
    return t, freq, timefreq, rms


def temporal_envelope(s, sample_rate, cutoff_freq=200.0, resample_rate=None):
    """
        Get the temporal envelope from the sound pressure waveform.

        s: the signal
        sample_rate: the sample rate of the signal
        cutoff_freq: the cutoff frequency of the low pass filter used to create the envelope

        Returns the temporal envelope of the signal, with same sample rate or downsampled.
    """

    #rectify a zeroed version
    srect = np.abs(s - np.mean(s))
    #low pass filter
    if cutoff_freq is not None:
        srect = lowpass_filter(srect, sample_rate, cutoff_freq, filter_order=4)
        srect[srect < 0] = 0
        

        
    if resample_rate is not None:
        lensound = len(srect)
        t=(np.array(range(lensound),dtype=float))/sample_rate
        lenresampled = int(round(float(lensound)*resample_rate/sample_rate))
        (srectresampled, tresampled) = resample(srect, lenresampled, t=t, axis=0, window=None)
        return (srectresampled, tresampled)
    else:   
        return srect


def recursive_ls(root_dir, file_pattern):
    """
        Walks through all the files in root_dir and returns every file whose name matches
        the pattern specified by file_pattern.
    """

    matches = list()
    for root, dirnames, filenames in os.walk(root_dir):
      for filename in fnmatch.filter(filenames, file_pattern):
          matches.append(os.path.join(root, filename))
    return matches


def sox_convert_to_mono(file_path):
    """
        Uses Sox (sox.sourceforge.net) to convert a stereo .wav file to mono.
    """

    root_dir,file_name = os.path.split(file_path)

    base_file_name = file_name[:-4]
    output_file_path = os.path.join(root_dir, '%s_mono.wav' % base_file_name)
    cmd = 'sox \"%s\" -c 1 \"%s\"' % (file_path, output_file_path)
    print(cmd)
    subprocess.call(cmd, shell=True)


def generate_sine_wave(duration, freq, samprate):
    """
        Generate a pure tone at a given frequency and sample rate for a specified duration.
    """

    t = np.arange(0.0, duration, 1.0 / samprate)
    return np.sin(2*np.pi*freq*t)


def generate_simple_stack(duration, fundamental_freq, samprate, num_harmonics=10):
    nsamps = int(duration*samprate)
    s = np.zeros(nsamps, dtype='float')
    ffreq = 0.0
    for n in range(num_harmonics):
        ffreq += fundamental_freq
        s += generate_sine_wave(duration, ffreq, samprate)
    return s


def generate_harmonic_stack(duration, fundamental_freq, samprate, num_harmonics=10, base=2):

    nsamps = int(duration*samprate)
    s = np.zeros(nsamps, dtype='float')
    for n in range(num_harmonics):
        freq = fundamental_freq * base**n
        s += generate_sine_wave(duration, freq, samprate)
    return s


def modulate_wave(s, samprate, freq):

    t = np.arange(len(s), dtype='float') / samprate
    c = np.sin(2*np.pi*t*freq)
    return c*s


def mtfft(spectrogram, df, dt, Log=False):
    """
        Compute the 2d modulation power and phase for a given time frequency slice.
        return temporal_freq,spectral_freq,mps_pow,mps_phase
    """

    #take the 2D FFT and center it
    smps = fft2(spectrogram)
    smps = fftshift(smps)

    #compute the log amplitude
    mps_pow = np.abs(smps)**2
    if Log:
        mps_pow = 10*np.log10(mps_pow)

    #compute the phase
    mps_phase = np.angle(smps)

    #compute the axes
    nf = mps_pow.shape[0]
    nt = mps_pow.shape[1]

    spectral_freq = fftshift(fftfreq(nf, d=df[1]-df[0]))
    temporal_freq = fftshift(fftfreq(nt, d=dt[1]-dt[0]))

    """
    nb = sdata.shape[1]
    dwf = np.zeros(nb)
    for ib in range(int(np.ceil((nb+1)/2.0))+1):
        posindx = ib
        negindx = nb-ib+2
        print 'ib=%d, posindx=%d, negindx=%d'% (ib, posindx, negindx )
        dwf[ib]= (ib-1)*(1.0/(df*nb))
        if ib > 1:
            dwf[negindx] =- dwf[ib]

    nt = sdata.shape[0]
    dwt = np.zeros(nt)
    for it in range(0, int(np.ceil((nt+1)/2.0))+1):
        posindx = it
        negindx = nt-it+2
        print 'it=%d, posindx=%d, negindx=%d' % (it, posindx, negindx)
        dwt[it] = (it-1)*(1.0/(nt*dt))
        if it > 1 :
            dwt[negindx] = -dwt[it]

    spectral_freq = dwf
    temporal_freq = dwt
    """

    return spectral_freq, temporal_freq, mps_pow, mps_phase

def mps(spectrogram, df, dt, window=None, Norm=True):
    """
    Calculates the modulation power spectrum using overlapp and add method with a gaussian window of length window in s
    Assumes that spectrogram is in dB.  df and dt are the axis of spectrogram.
    """
    
    # Debugging paramenter
    debugPlot = False

    # Resolution of spectrogram in DB
    dbRES = 50 
    
    # Check the size of the spectrogram vs dt
    nt = dt.size
    nf = df.size
    if spectrogram.shape[1] != nt and spectrogram.shape[0] != nf:   
        print('Error in mps. Expected  %d bands in frequency and %d points in time' % (nf, nt))
        print('Spectrogram had shape %d, %d' % spectrogram.shape)
        return 0, 0, 0
        
    # Z-score the flattened spectrogram is Norm is True
    sdata = deepcopy(spectrogram)
    
    if Norm:
        maxdata = sdata.max()
        mindata = maxdata - dbRES
        sdata[sdata< mindata] = mindata
        sdata -= sdata.mean()
        sdata /= sdata.std()
        
    if window == None:
        window = dt[-1]/10.0
            
    # Find the number of spectrogram points in the gaussian window 
    if dt[-1] < window:
        print('Warning in mps: Requested MPS window size is greater than spectrogram temporal extent.')
        print('mps will be calculate with windows of %d points or %s s' % (nt-1, dt[-1]) )
        nWindow = nt - 1
    else:
        nWindow = np.where(dt>= window)[0][0]
    if nWindow%2 == 0:
        nWindow += 1  # Make it odd size so that we have a symmetric window
        
    if nWindow < 64:
        print('Error in mps: window size %d pts (%.3f s) is two small for reasonable estimates' % (nWindow, window))
        return np.asarray([]), np.asarray([]), np.asarray([])
        
    # Generate the Gaussian window
    gt, wg = gaussian_window(nWindow, 6)
    tShift = int(gt[-1]/3)
    nchunks = 0
    
    # Pad the spectrogram with zeros.
    minSdata = sdata.min()
    sdataZeros = np.ones((sdata.shape[0], int((nWindow-1)/2))) * minSdata
    sdata = np.concatenate((sdataZeros, sdata, sdataZeros), axis = 1)
    
    if debugPlot:
        plt.figure(1)
        plt.clf()
        plt.subplot()
        plt.imshow(sdata, origin='lower')
        plt.title('Scaled and Padded Spectrogram')
        plt.show()
        plt.pause(1)

    
    for tmid in range(tShift, nt, tShift):
        
        # t mid is in the original coordinates while tstart and tend
        # are shifted to deal with the zero padding.
        tstart = tmid-(nWindow-1)//2-1
        tstart += (nWindow-1)//2
        if tstart < 0:
            print('Error in mps. tstart negative')
            break;
                       
        tend = tmid+(nWindow-1)//2
        tend += (nWindow-1)//2
        if tend > sdata.shape[1]:
            print('Error in mps. tend too large')
            break
        nchunks += 1
        
        # Multiply the spectrogram by the window
        wSpect = deepcopy(sdata[:,tstart:tend])
        
                # Debugging code
        if debugPlot:
            plt.figure(nchunks+1)
            plt.clf()
            plt.subplot(121)
            plt.imshow(wSpect, origin='lower')
            plt.title('%d Middle (%d %d)' % (tmid, tstart, tend) )

            
        for fInd in range(nf):
            wSpect[fInd,:] = wSpect[fInd,:]*wg
            
        # Debugging code
        if debugPlot:
            plt.figure(nchunks+1)
            plt.subplot(122)
            plt.imshow(wSpect, origin='lower')
            plt.title('After')
            plt.show()
            plt.pause(1)
            input("Press Enter to continue...")
            
        # Get the 2d FFT
        wf, wt, mps_pow, mps_phase = mtfft(wSpect, df, dt[0:tend-tstart])
        if nchunks == 1:
            mps_powAvg = mps_pow
        else:
            mps_powAvg += mps_pow
            
    mps_powAvg /= nchunks
    
    return wf, wt, mps_powAvg


def plot_mps(spectral_freq, temporal_freq, amp, phase=None):

    plt.figure()

    #plot the amplitude
    if phase:
        plt.subplot(2, 1, 1)
        
    #ex = (spectral_freq.min(), spectral_freq.max(), temporal_freq.min(), temporal_freq.max())
    ex = (temporal_freq.min(), temporal_freq.max(), spectral_freq.min()*1e3, spectral_freq.max()*1e3)
    plt.imshow(amp, interpolation='nearest', aspect='auto', origin='lower', cmap=cmap.jet, extent=ex)
    plt.ylabel('Spectral Frequency (Cycles/KHz)')
    plt.xlabel('Temporal Frequency (Hz)')
    plt.colorbar()
    plt.ylim((0,spectral_freq.max()*1e3))
    plt.title('Power')

    #plot the phase
    if phase:
        plt.subplot(2, 1, 2)
        plt.imshow(phase, interpolation='nearest', aspect='auto', origin='lower', cmap=cmap.jet, extent=ex)
        plt.ylabel('Spectral Frequency (Cycles/KHz)')
        plt.xlabel('Temporal Frequency (Hz)')
        plt.ylim((0,spectral_freq.max()*1e3))
        plt.title('Phase')
        plt.colorbar()

def synSpect(b, x):
# Generates a model spectrum made out of gaussian peaks
# fund, sigma, pkmax, dbfloor
# global fundGlobal maxFund minFund

    npeaks = np.size(b)-1  # First element of b is the sampling rate
# amp = 25      # Force 25 dB peaks
    sdpk = 60    # Force 80 hz width

    synS = np.zeros(len(x))


    for i in range(npeaks):
        a = b[i+1]   # To inforce positive peaks only
        synS = synS + a*np.exp(-(x-b[0]*(i+1))**2/(2*sdpk**2))
    
    #if (sum(isinf(synS)) + sum(isnan(synS))):
    #    for i in range(npeaks):
    #        fprintf(1,'%f ', exp(b(i+1)))    
    return synS

    
def residualSyn(vars, x, realS):
    b = vars
    synS = synSpect(b, x)
    
    return realS-synS
    
    #if (sum(isinf(synS)) + sum(isnan(synS)))
    #    for i=1:npeaks
    #        fprintf(1,'%f ', exp(b(i+1)))  
    
def lpc(signal, order):
    """Compute the Linear Prediction Coefficients.

    Return the order + 1 LPC coefficients for the signal. c = lpc(x, k) will
    find the k+1 coefficients of a k order linear filter:

      xp[n] = -c[1] * x[n-2] - ... - c[k-1] * x[n-k-1]

    Such as the sum of the squared-error e[i] = xp[i] - x[i] is minimized.

    Parameters
    ----------
    signal: array_like
        input signal   
    order : int
        LPC order (the output will have order + 1 items)"""

    order = int(order)

    if signal.ndim > 1:
        raise ValueError("Array of rank > 1 not supported yet")
    if order > signal.size:
        raise ValueError("Input signal must have a lenght >= lpc order")

    if order > 0:
        p = order + 1
        r = np.zeros(p, signal.dtype)
        # Number of non zero values in autocorrelation one needs for p LPC
        # coefficients
        nx = np.min([p, signal.size])
        x = np.correlate(signal, signal, 'full')
        r[:nx] = x[signal.size-1:signal.size+order]
        phi = np.dot(inv(toeplitz(r[:-1])), -r[1:])
        return np.concatenate(([1.], phi)), None, None
    else:
        return np.ones(1, dtype = signal.dtype), None, None


def fundEstimator(soundIn, fs, t=None, debugFig = 0, maxFund = 1500, minFund = 300, lowFc = 200, highFc = 6000, minSaliency = 0.5, minFormantFreq = 500, maxFormantBW = 500, method='Stack'):
    """
    Estimates the fundamental frequency of a complex sound.
    soundIn is the sound pressure waveformlog spectrogram.
    fs is the sampling rate
    t is a vector of time values in s at which the fundamental will be estimated.
    The sound must include at least 1024 sample points

    The optional parameter with defaults are
    Some user parameters (should be part of the function at some time)
       debugFig = 0         Set to zero to eliminate figures.
       maxFund = 1500       Maximum fundamental frequency
       minFund = 300        Minimum fundamental frequency
       lowFc = 200          Low frequency cut-off for band-passing the signal prior to auto-correlation.
       highFc = 6000        High frequency cut-off
       minSaliency = 0.5    Threshold in the auto-correlation for minimum saliency - returns NaN for pitch values is saliency is below this number

    Four methods are available: 
    'AC' - Peak of the auto-correlation function
    'ACA' - Peak of envelope of auto-correlation function 
    'Cep' - First peak in cepstrum 
    'Stack' - Fitting of harmonic stacks (default - works well for zebra finches)
    
    Returns
           sal     - the time varying pitch saliency - a number between 0 and 1 corresponding to relative size of the first auto-correlation peak
           fund     - the time-varying fundamental in Hz at the same resolution as the spectrogram.
           fund2   - a second peak in the spectrum - not a multiple of the fundamental a sign of a second voice
           form1   - the first formant, if it exists
           form2   - the second formant, if it exists
           form3   - the third formant, if it exists
           soundLen - length of sal, fund, fund2, form1, form2, form3
    """

    # Band-pass filtering signal prior to auto-correlation
    soundLen = len(soundIn)
    nfilt = 1024
    if soundLen < 1024:
        print('Error in fundEstimator: sound too short for bandpass filtering, len(soundIn)=%d' % soundLen)
        return (np.asarray([]), np.asarray([]), np.asarray([]), np.asarray([]), np.asarray([]), np.asarray([]), soundLen)

    # high pass filter the signal
    highpassFilter = firwin(nfilt-1, 2.0*lowFc/fs, pass_zero=False)
    padlen = min(soundLen-10, 3*len(highpassFilter))
    soundIn = filtfilt(highpassFilter, [1.0], soundIn, padlen=padlen)

    # low pass filter the signal
    lowpassFilter = firwin(nfilt, 2.0*highFc/fs)
    padlen = min(soundLen-10, 3*len(lowpassFilter))
    soundIn = filtfilt(lowpassFilter, [1.0], soundIn, padlen=padlen)

    # Plot a spectrogram?
    #if debugFig:
    #    plt.figure(9)
    #    (tDebug ,freqDebug ,specDebug , rms) = spectrogram(soundIn, fs, 1000.0, 50, min_freq=0, max_freq=10000, nstd=6, log=True, noise_level_db=50, rectify=True) 
    #    plot_spectrogram(tDebug, freqDebug, specDebug)

    # Initializations and useful variables
    soundLen = len(soundIn)
    sound_dur = soundLen / fs
    
    if t is None:
        # initialize t to be spaced by  1 ms increments if not specified     
        _si = 1e-3
        npts = int(sound_dur / _si)
        t = np.arange(npts) * _si

    nt=len(t)
    soundRMS = np.zeros(nt)
    fund = np.zeros(nt)
    fund2 = np.zeros(nt)
    sal = np.zeros(nt)
    form1 = np.zeros(nt)
    form2 = np.zeros(nt)
    form3 = np.zeros(nt)

    #  Calculate the size of the window for the auto-correlation
    alpha = 5                          # Number of sd in the Gaussian window
    winLen = int(np.fix((2.0*alpha/minFund)*fs))  # Length of Gaussian window based on minFund
    if (winLen%2 == 0):  # Make a symmetric window
        winLen += 1
        
    winLen2 = 2**12+1   # This looks like a good size for LPC - 4097 points

    gt, w = gaussian_window(winLen, alpha)
    gt2, w2 = gaussian_window(winLen2, alpha)
    maxlags = int(2*ceil((float(fs)/minFund)))

    # First calculate the rms in each window
    for it in range(nt):
        tval = t[it]               # Center of window in time
        if tval >= sound_dur:
            continue
        tind = int(np.fix(tval*fs))    # Center of window in ind
        tstart = tind - (winLen-1)//2
        tend = tind + (winLen-1)//2
    
        if tstart < 0:
            winstart = - tstart
            tstart = 0
        else:
            winstart = 0
        
        if tend >= soundLen:
            windend = winLen - (tend-soundLen+1) - 1
            tend = soundLen-1
        else:
            windend = winLen-1
            
        soundWin = soundIn[tstart:tend]*w[winstart:windend]
        soundRMS[it] = np.std(soundWin)
    
    soundRMSMax = max(soundRMS)

    # Calculate the auto-correlation in windowed segments and obtain 4 guess values of the fundamental
    # fundCorrGuess - guess from the auto-correlation function
    # fundCorrAmpGuess - guess form the amplitude of the auto-correlation function
    # fundCepGuess - guess from the cepstrum
    # fundStackGuess - guess taken from a fit of the power spectrum with a harmonic stack, using the fundCepGuess as a starting point
    #  Current version use fundStackGuess as the best estimate...

    soundlen = 0
    for it in range(nt):
        fund[it] = float('nan')
        sal[it] = float('nan')
        fund2[it] = float('nan')
        form1[it] = float('nan')
        form2[it] = float('nan')
        form3[it] = float('nan')
        
        if (soundRMS[it] < soundRMSMax*0.1):
            continue
    
        soundlen += 1
        tval = t[it]               # Center of window in time
        if tval >= sound_dur:       # This should not happen here because the RMS should be zero
            continue
        tind = int(np.fix(tval*fs))    # Center of window in ind
        tstart = tind - (winLen-1)//2
        tend = tind + (winLen-1)//2
    
        if tstart < 0:
            winstart = - tstart
            tstart = 0
        else:
            winstart = 0
        
        if tend >= soundLen:
            windend = winLen - (tend-soundLen+1) - 1
            tend = soundLen-1
        else:
            windend = winLen-1
            
        tstart2 = tind - (winLen2-1)//2
        tend2 = tind + (winLen2-1)//2
    
        if tstart2 < 0:
            winstart2 = - tstart2
            tstart2 = 0
        else:
            winstart2 = 0
        
        if tend2 >= soundLen:
            windend2 = winLen2 - (tend2-soundLen+1) - 1
            tend2 = soundLen-1
        else:
            windend2 = winLen2-1
            
        soundWin = soundIn[tstart:tend]*w[winstart:windend]
              
        soundWin2 = soundIn[tstart2:tend2]*w2[winstart2:windend2]
        
        # Apply LPC to get time-varying formants and one additional guess for the fundamental frequency
        # TODO (kevin): replace this with librosa 
        A, E, K = lpc(soundWin2, 8)    # 8 degree polynomial
        rts = np.roots(A)          # Find the roots of A
        rts = rts[np.imag(rts)>=0]  # Keep only half of them
        angz = np.arctan2(np.imag(rts),np.real(rts))
    
        # Calculate the frequencies and the bandwidth of the formants
        frqsFormants = angz*(fs/(2*np.pi))
        indices = np.argsort(frqsFormants)
        bw = -0.5*(fs/(2*np.pi))*np.log(np.abs(rts))  # FIXME (kevin): I think this line was broken before... it was using 1/2
    
        # Keep formants above 500 Hz and with bandwidth < 500 # This was 1000 for bird calls
        formants = []
        for kk in indices:
            if ( frqsFormants[kk]>minFormantFreq and bw[kk] < maxFormantBW):        
                formants.append(frqsFormants[kk])
        formants = np.array(formants) 
        
        if len(formants) > 0 : 
            form1[it] = formants[0]
        if len(formants) > 1 : 
            form2[it] = formants[1]
        if len(formants) > 2 : 
            form3[it] = formants[2]

        # Calculate the auto-correlation
        lags = np.arange(-maxlags, maxlags+1, 1)
        autoCorr = correlation_function(soundWin, soundWin, lags)
        ind0 = int(np.where(lags == 0)[0][0])  # need to find lag zero index
    
        # find peaks
        indPeaksCorr = detect_peaks(autoCorr, mph=autoCorr.max()/10.0)
    
        # Eliminate center peak and all peaks too close to middle    
        indPeaksCorr = np.delete(indPeaksCorr,np.where( (indPeaksCorr-ind0) < fs/maxFund)[0])
        pksCorr = autoCorr[indPeaksCorr]
    
        # Find max peak
        if len(pksCorr)==0:
            pitchSaliency = 0.1               # 0.1 goes with the detection of peaks greater than max/10
        else:
            indIndMax = np.where(pksCorr == max(pksCorr))[0][0]
            indMax = indPeaksCorr[indIndMax]   
            fundCorrGuess = fs/abs(lags[indMax])
            pitchSaliency = autoCorr[indMax]/autoCorr[ind0]

        sal[it] = pitchSaliency
    
        if sal[it] < minSaliency:
            continue

        # Calculate the envelope of the auto-correlation after rectification
        envCorr = temporal_envelope(autoCorr, fs, cutoff_freq=maxFund, resample_rate=None) 
        locsEnvCorr = detect_peaks(envCorr, mph=envCorr.max()/10.0)
        pksEnvCorr = envCorr[locsEnvCorr]
    
        # Find the peak closest to zero
        if locsEnvCorr.size > 1:
            lagdiff = np.abs(locsEnvCorr[0]-ind0)
            indIndEnvMax = 0
                             
            for indtest in range(1,locsEnvCorr.size):
                lagtest = np.abs(locsEnvCorr[indtest]-ind0)
                if lagtest < lagdiff:
                    lagdiff = lagtest
                    indIndEnvMax = indtest                   
          
        # Take the first peak after the one closest to zero
            if indIndEnvMax+2 > len(locsEnvCorr):   # No such peak - use data for correlation function
                fundCorrAmpGuess = fundCorrGuess
                indEnvMax = indMax
            else:
                indEnvMax = locsEnvCorr[indIndEnvMax+1]
                if lags[indEnvMax] == 0 :  # This should not happen
                    print('Error: Max Peak in enveloppe auto-correlation found at zero delay')
                    fundCorrAmpGuess = fundCorrGuess
                    indEnvMax = indMax
                else:
                    fundCorrAmpGuess = fs/lags[indEnvMax]
        else:
            fundCorrAmpGuess = fundCorrGuess
            indEnvMax = indMax

        # Calculate power spectrum and cepstrum
        Y = fft(soundWin, n=winLen+1)
        f = (fs/2.0)*(np.array(range(int((winLen+1)/2+1)), dtype=float)/float((winLen+1)//2))
        fhigh = np.where(f >= highFc)[0][0]
    
        powSound = 20.0*np.log10(np.abs(Y[0:(winLen+1)//2+1]))    # This is the power spectrum
        powSoundGood = powSound[0:fhigh]
        maxPow = max(powSoundGood)
        powSoundGood = powSoundGood - maxPow   # Set zero as the peak amplitude
        powSoundGood[powSoundGood < - 60] = -60    
    
        # Calculate coarse spectral enveloppe
        p = np.polyfit(f[0:fhigh], powSoundGood, 3)
        powAmp = np.polyval(p, f[0:fhigh]) 
    
        # Cepstrum
        CY = dct(powSoundGood-powAmp, norm = 'ortho')            
    
        tCY = 1000.0*np.array(range(len(CY)))/fs          # Units of Cepstrum in ms
        fCY = np.zeros(tCY.size)
        fCY[1:] = 1000.0/tCY[1:] # Corresponding fundamental frequency in Hz.
        fCY[0] = fs*2.0          # Nyquist limit not infinity
        lowInd = np.where(fCY<lowFc)[0]
        if lowInd.size > 0:
            flowCY = np.where(fCY < lowFc)[0][0]
        else:
            flowCY = fCY.size
            
        fhighCY = np.where(fCY < highFc)[0][0]
    
        # Find peak of Cepstrum
        indPk = np.where(CY[fhighCY:flowCY] == max(CY[fhighCY:flowCY]))[0][-1]
        indPk = fhighCY + indPk 
        fmass = 0
        mass = 0
        indTry = indPk
        while (CY[indTry] > 0):
            fmass = fmass + fCY[indTry]*CY[indTry]
            mass = mass + CY[indTry]
            indTry = indTry + 1
            if indTry >= len(CY):
                break

        indTry = indPk - 1
        if (indTry >= 0 ):
            while (CY[indTry] > 0):
                fmass = fmass + fCY[indTry]*CY[indTry]
                mass = mass + CY[indTry]
                indTry = indTry - 1
                if indTry < 0:
                    break

        fGuess = fmass/mass
    
        if (fGuess == 0  or np.isnan(fGuess) or np.isinf(fGuess) ):              # Failure of cepstral method
            fGuess = fundCorrGuess

        fundCepGuess = fGuess
    
        # Force fundamendal to be bounded
        if (fundCepGuess > maxFund ):
            i = 2
            while(fundCepGuess > maxFund):
                fundCepGuess = fGuess/i
                i += 1
        elif (fundCepGuess < minFund):
            i = 2
            while(fundCepGuess < minFund):
                fundCepGuess = fGuess*i
                i += 1
    
        # Fit Gaussian harmonic stack
        maxPow = max(powSoundGood-powAmp)

        # This is the matlab code...
        # fundFitCep = NonLinearModel.fit(f(1:fhigh)', powSoundGood'-powAmp, @synSpect, [fundCepGuess ones(1,9).*log(maxPow)])
        # modelPowCep = synSpect(double(fundFitCep.Coefficients(:,1)), f(1:fhigh))

        vars = np.concatenate(([fundCorrGuess], np.ones(9)*np.log(maxPow)))
        bout = leastsq(residualSyn, vars, args = (f[0:fhigh], powSoundGood-powAmp)) 
        modelPowCep = synSpect(bout[0], f[0:fhigh])
        errCep = sum((powSoundGood - powAmp - modelPowCep)**2)
    
        vars = np.concatenate(([fundCorrGuess*2], np.ones(9)*np.log(maxPow)))
        bout2 = leastsq(residualSyn, vars, args = (f[0:fhigh], powSoundGood-powAmp)) 
        modelPowCep2 = synSpect(bout2[0], f[0:fhigh])
        errCep2 = sum((powSoundGood - powAmp - modelPowCep2)**2)
    
        if errCep2 < errCep:
            bout = bout2
            modelPowCep =  modelPowCep2

        fundStackGuess = bout[0][0]
        if (fundStackGuess > maxFund) or (fundStackGuess < minFund ):
            fundStackGuess = float('nan')

        # Store the result depending on the method chosen
        if method == 'AC':
            fund[it] = fundCorrGuess
        elif method == 'ACA':
            fund[it] = fundCorrAmpGuess
        elif method == 'Cep':
            fund[it] = fundCepGuess
        elif method == 'Stack':
            fund[it] = fundStackGuess  
            
        # A second cepstrum for the second voice
        #     CY2 = dct(powSoundGood-powAmp'- modelPowCep)
    
        if  not np.isnan(fundStackGuess):
            powLeft = powSoundGood- powAmp - modelPowCep
            maxPow2 = max(powLeft)
            f2 = 0
            if ( maxPow2 > maxPow*0.5):    # Possible second peak in central area as indicator of second voice.
                f2 = f[np.where(powLeft == maxPow2)[0][0]]
                if ( f2 > 1000 and f2 < 4000):
                    if (pitchSaliency > minSaliency):
                        fund2[it] = f2

#%     modelPowCorrAmp = synSpect(double(fundFitCorrAmp.Coefficients(:,1)), f(1:fhigh))
#%     
#%     errCorr = sum((powSoundGood - powAmp' - modelPowCorr).^2)
#%     errCorrAmp = sum((powSoundGood - powAmp' - modelPowCorrAmp).^2)
#%     errCorrSum = sum((powSoundGood - powAmp' - (modelPowCorr+modelPowCorrAmp) ).^2)
#%       
#%     f1 = double(fundFitCorr.Coefficients(1,1))
#%     f2 = double(fundFitCorrAmp.Coefficients(1,1))
#%     
#%     if (pitchSaliency > minSaliency)
#%         if (errCorr < errCorrAmp)
#%             fund(it) = f1
#%             if errCorrSum < errCorr
#%                 fund2(it) = f2
#%             end
#%         else
#%             fund(it) = f2
#%             if errCorrSum < errCorrAmp
#%                 fund2(it) = f1
#%             end
#%         end
#%         
#%     end

        if (debugFig ):
            plt.figure(10)
            plt.subplot(4,1,1)
            plt.cla()
            plt.plot(soundWin)
#         f1 = double(fundFitCorr.Coefficients(1,1))
#         f2 = double(fundFitCorrAmp.Coefficients(1,1))
            titleStr = 'Saliency = %.2f F0 AC = %.2f ACA = %.2f Cep = %.2f St = %.2f(Hz)' % (pitchSaliency, fundCorrGuess, fundCorrAmpGuess, fundCepGuess, fundStackGuess)
            plt.title(titleStr)
        
            plt.subplot(4,1,2)
            plt.cla()
            plt.plot(1000*(lags/fs), autoCorr)
            plt.plot([1000.*lags[indMax]/fs, 1000.*lags[indMax]/fs], [0, autoCorr[ind0]], 'k')
            plt.plot(1000.*lags/fs, envCorr, 'r', linewidth= 2)
            plt.plot([1000*lags[indEnvMax]/fs, 1000.*lags[indEnvMax]/fs], [0, autoCorr[ind0]], 'g', linewidth=2)
            plt.xlabel('Time (ms)')
              
            plt.subplot(4,1,3)
            plt.cla()
            plt.plot(f[0:fhigh],powSoundGood)
            plt.axis([0, highFc, -60, 0])
            plt.plot(f[0:fhigh], powAmp, 'b--')
            plt.plot(f[0:fhigh], modelPowCep + powAmp, 'k')
            # plt.plot(f(1:fhigh), modelPowCorrAmp + powAmp', 'g')
        
            for ih in range(1,6):
                plt.plot([fundCorrGuess*ih, fundCorrGuess*ih], [-60, 0], 'k')
                plt.plot([fundCorrAmpGuess*ih, fundCorrAmpGuess*ih], [-60, 0], 'g')
                plt.plot([fundStackGuess*ih, fundStackGuess*ih], [-60, 0], 'r')

                plt.plot([fundCepGuess*ih, fundCepGuess*ih], [-60, 0], 'y')

            if f2 != 0: 
                plt.plot([f2, f2], [-60, 0], 'b')

            plt.xlabel('Frequency (Hz)')
            # title(sprintf('Err1 = %.1f Err2 = %.1f', errCorr, errCorrAmp))
        
            plt.subplot(4,1,4)
            plt.cla()
            plt.plot(tCY, CY)
#         plot(tCY, CY2, 'k--')
            plt.plot([1000./fundCorrGuess, 1000./fundCorrGuess], [0, max(CY)], 'k')
            plt.plot([1000./fundCorrAmpGuess, 1000./fundCorrAmpGuess], [0, max(CY)], 'g')
            plt.plot([1000./fundStackGuess, 1000./fundStackGuess], [0, max(CY)], 'r')
            plt.plot([1000./fundCepGuess, 1000./fundCepGuess], [0, max(CY)], 'y')
        
            #%         plot([(pkClosest-1)/fs (pkClosest-1)/fs], [0 max(CY)], 'g')
            #%         if ~isempty(ipk2)
            #%             plot([(pk2-1)/fs (pk2-1)/fs], [0 max(CY)], 'b')
            #%         end
            #%         for ip=1:length(pks)
            #%             plot([(locs(ip)-1)/fs (locs(ip)-1)/fs], [0 pks(ip)/4], 'r')
            #%         end
            plt.axis([0, 1000.*np.size(CY)/(2.*fs), 0, max(CY)])
            plt.xlabel('Time (ms)')

            plt.pause(1)
    
    # Fix formants.
    meanf1 = np.mean(form1[~np.isnan(form1)])
    meanf2 = np.mean(form2[~np.isnan(form2)])
    meanf3 = np.mean(form3[~np.isnan(form3)])

    for it in range(nt):
        if ~np.isnan(form1[it]):
            df11 = np.abs(form1[it]-meanf1)
            df12 = np.abs(form1[it]-meanf2)
            df13 = np.abs(form1[it]-meanf3)
            if df12 < df11:
                if df13 < df12:
                    if ~np.isnan(form3[it]):
                        df33 = np.abs(form3[it]-meanf3)
                        if df13 < df33:
                            form3[it] = form1[it]
                    else:
                      form3[it] = form1[it]
                else:
                    if ~np.isnan(form2[it]):
                        df22 = np.abs(form2[it]-meanf2)
                        if df12 < df22:
                            form2[it] = form1[it]
                    else:
                        form2[it] = form1[it]
                form1[it] = float('nan')
            if ~np.isnan(form2[it]):  
                df21 = np.abs(form2[it]-meanf1)
                df22 = np.abs(form2[it]-meanf2)
                df23 = np.abs(form2[it]-meanf3)
                if df21 < df22 :
                    if ~np.isnan(form1[it]):
                        df11 = np.abs(form1[it]-meanf1)
                        if df21 < df11:
                            form1[it] = form2[it]
                    else:
                      form1[it] = form2[it]
                    form2[it] = float('nan')
                elif df23 < df22:
                    if ~np.isnan(form3[it]):
                        df33 = np.abs(form3[it]-meanf3)
                        if df23 < df33:
                            form3[it] = form2[it]
                    else:
                        form3[it] = form2[it]
                    form2[it] = float('nan')
            if ~np.isnan(form3[it]):
                df31 = np.abs(form3[it]-meanf1)
                df32 = np.abs(form3[it]-meanf2)
                df33 = np.abs(form3[it]-meanf3)
                if df32 < df33:
                    if df31 < df32:
                        if ~np.isnan(form1[it]):
                            df11 = np.abs(form1[it]-meanf1)
                            if df31 < df11:
                                form1[it] = form3[it]
                        else:
                            form1[it] = form3[it]
                    else:
                        if ~np.isnan(form2[it]):
                            df22 = np.abs(form2[it]-meanf2)
                            if df32 < df22:
                                form2[it] = form3[it]
                        else:
                            form2[it] = form3[it]
                    form3[it] = float('nan')

    return (sal, fund, fund2, form1, form2, form3, soundlen)



def get_mps(t, freq, spec):
    "Computes the MPS of a spectrogram (idealy a log-spectrogram) or other REAL time-freq representation"
    mps = fftshift(fft2(spec))
    amps = np.real(mps * np.conj(mps))
    nf = mps.shape[0]
    nt = mps.shape[1]
    wfreq = fftshift(fftfreq(nf, d=freq[1] - freq[0]))
    wt = fftshift(fftfreq(nt, d=t[1] - t[0]))
    return wt, wfreq, mps, amps

def inverse_mps(mps):
    "Inverts a MPS back to a spectrogram"
    spec = ifft2(ifftshift(mps))
    return spec



def play_signal(s, normalize = False):
    "quick and easy temporary play"
    wf = WavFile()
    wf.sample_rate = 44100 #standard samp rate
    wf.data = s
    wf.to_wav("/tmp/README.wav", normalize)
    play_sound("/tmp/README.wav")



def inverse_spectrogram(spec, s_len,
    sample_rate, spec_sample_rate, freq_spacing, min_freq=0, max_freq=None, nstd=6, log=True, noise_level_db=80, rectify=True):
    """turns the complex spectrogram into a signal

    inverts by repeating the process on a string-of-ones
    """

    spec_copy = spec.copy()
    if log:
        spec_copy = 10**(spec_copy)
    spec_tranpose = spec.transpose() # spec_tranpose[time][frequency]

    hnwinlen = len(spec) - 1
    nincrement = int(np.round(float(sample_rate)/spec_sample_rate))

    gauss_t = np.arange(-hnwinlen, hnwinlen+1, 1.0)
    gauss_std = float(2*hnwinlen) / float(nstd)
    gauss_window = np.exp(-gauss_t**2 / (2.0*gauss_std**2)) / (gauss_std*np.sqrt(2*np.pi))
    
    s = np.zeros(s_len + 2*hnwinlen+1)
    w = np.zeros(s_len + 2*hnwinlen+1)

    for i in range(len(spec_tranpose)):
        sample = i * nincrement
        spec_slice = np.concatenate((spec_tranpose[i][:0:-1].conj(), spec_tranpose[i]))
        s[sample:sample+2*hnwinlen+1] += gauss_window * ifft(ifftshift(spec_slice))
        w[sample:sample+2*hnwinlen+1] += gauss_window ** 2
    s /= w
    return s[hnwinlen:hnwinlen+s_len]


def inverse_real_spectrogram(spec, s_len,
    sample_rate, spec_sample_rate, freq_spacing, min_freq=0, max_freq=None, nstd=6, log=True, noise_level_db=80, rectify=True, iterations = 10):
    "inverts a real spectrogram into a signal using the griffith/lim algorithm"
    spec_magnitude = spec.copy()

    if log:
        spec_magnitude = 10**spec_magnitude
    estimated = inverse_spectrogram(spec_magnitude, s_len, sample_rate, spec_sample_rate, freq_spacing, min_freq, max_freq, nstd, log=False)
    for i in range(iterations):
        phase_spec = spectrogram(estimated, sample_rate, spec_sample_rate, freq_spacing, min_freq, max_freq, nstd, log=False)[2]
        error = ((abs(spec_magnitude) - abs(phase_spec))**2).sum() / (abs(spec_magnitude)**2).sum()
        print('the error after iteration %d is %f' % (i+i, error))
        spec_angle = np.angle(phase_spec)
        estimated_spec = spec_magnitude * np.exp(1j*spec_angle)
        estimated = inverse_spectrogram(estimated_spec, s_len, sample_rate, spec_sample_rate, freq_spacing, min_freq, max_freq, nstd, log=False)
    return estimated


def log_transform(x, dbnoise=100, normalize=False):
    """ Takes the log of a power spectrum or spectrogram to convert into decibels.

    :param x: The power spectrum or spectrogram. The contents of x will be replaced with the log version.
    :param dbnoise: The noise level in decibels. Anything lower than dbnoise will be set to zero.
    """
    x /= x.max()
    zi = x > 0
    x[zi] = 20*np.log10(x[zi]) + dbnoise
    x[x < 0] = 0
    if normalize:
        x /= x.max()


def spec_stats(spec_t, spec_freq, spec):
    """ Compute time-varying statistics on a spectrogram (or log spectrogram).

    :param spec_t: Spectrogram times with shape (num_time_points)
    :param spec_freq: Spectrogram frequencies with shape (num_freq)
    :param spec: Spectrogram of shape (num_freq, num_time_points)

    :return:
    """

    # normalize each time point by it's sum to create a probability distribution
    nfreq,nt = spec.shape
    spec_p = deepcopy(spec)
    spec_p -= spec_p.min()
    spec_p_sum = spec_p.sum(axis=0)
    spec_p /= spec_p_sum

    # compute mean frequency
    freq_mean = np.dot(spec_p.T, spec_freq)

    # compute quantiles
    spec_p_csum = np.cumsum(spec_p, axis=0)

    qvals = [0.25, 0.5, 0.75]
    Q = np.zeros([len(qvals), nt])
    for t in range(nt):
        for k,q in enumerate(qvals):
            i = spec_p_csum[:, t] <= q
            if i.sum() > 0:
                fi = np.max(np.where(i)[0])
                Q[k, t] = spec_freq[fi]

    stats = dict()
    stats['Q'] = Q
    stats['qvals'] = qvals
    stats['freq_mean'] = freq_mean

    return stats





import numpy as np
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis as QDA
from sklearn.ensemble import RandomForestClassifier as RF
from sklearn.model_selection import StratifiedKFold
from scipy.stats import binom


def discriminatePlot(X, y, cVal, titleStr='', figdir='.', Xcolname = None, plotFig = False, removeTickLabels = False, testInd = None):
    # Frederic's Robust Wrapper for discriminant analysis function.  Performs lda, qda and RF afer error checking, 
    # Generates nice plots and returns cross-validated
    # performance, stderr and base line.
    # X np array n rows x p parameters
    # y group labels n rows
    # rgb color code for each data point - should be the same for each data beloging to the same group
    # titleStr title for plots
    # figdir is a directory name (folder name) for figures
    # Xcolname is a np.array or list of strings with column names for printout display
    # returns: ldaScore, ldaScoreSE, qdaScore, qdaScoreSE, rfScore, rfScoreSE, nClasses

    # Global Parameters
    CVFOLDS = 10
    MINCOUNT = 10
    MINCOUNTTRAINING = 5
    # figdir = '/Users/frederictheunissen/Documents/Data/Julie/Acoustical Analysis/Figures Voice'

    # Initialize Variables and clean up data
    classes, classesCount = np.unique(y, return_counts = True)  # Classes to be discriminated should be same as ldaMod.classes_
    goodIndClasses = np.array([n >= MINCOUNT for n in classesCount])
    goodInd = np.array([b in classes[goodIndClasses] for b in y])
    if testInd is not None:
        # Check for goodInd - should be an np.array of dtype=bool
        # Transform testInd into an index inside xGood and yGood
        testIndx = testInd.nonzero()[0]
        goodIndx = goodInd.nonzero()[0]
        testInd = np.hstack([ np.where(goodIndx == testval)[0] for testval in testIndx])
        trainInd = np.asarray([i for i in range(len(goodIndx)) if i not in testInd])
        
    yGood = y[goodInd]
    XGood = X[goodInd]
    cValGood = cVal[goodInd]
        
    classes, classesCount = np.unique(yGood, return_counts = True) 
    nClasses = classes.size         # Number of classes or groups  

    # Do we have enough data?  
    if (nClasses < 2):
        print ('Error in ldaPLot: Insufficient classes with minimun data (%d) for discrimination analysis' % (MINCOUNT))
        return -1, -1, -1, -1 , -1, -1, -1, -1, -1
    
    if testInd is None:
        cvFolds = min(min(classesCount), CVFOLDS)
        if (cvFolds < CVFOLDS):
            print ('Warning in ldaPlot: Cross-validation performed with %d folds (instead of %d)' % (cvFolds, CVFOLDS))
    else:
        cvFolds = 1
   
    # Data size and color values   
    nD = XGood.shape[1]                 # number of features in X
    nX = XGood.shape[0]                 # number of data points in X
    cClasses = []   # Color code for each class
    for cl in classes:
        icl = (yGood == cl).nonzero()[0][0]
        cClasses.append(np.append(cValGood[icl],1.0))
    cClasses = np.asarray(cClasses)
    
    # Use a uniform prior 
    myPrior = np.ones(nClasses)*(1.0/nClasses)  

    # Perform a PCA for dimensionality reduction so that the covariance matrix can be fitted.
    nDmax = int(np.fix(np.sqrt(nX//5)))
    if nDmax < nD:
        print ('Warning: Insufficient data for', nD, 'parameters. PCA projection to', nDmax, 'dimensions.' )
    nDmax = min(nD, nDmax)
    pca = PCA(n_components=nDmax)
    Xr = pca.fit_transform(XGood)
    print ('Variance explained is %.2f%%' % (sum(pca.explained_variance_ratio_)*100.0))
    
    
    # Initialise Classifiers  
    ldaMod = LDA(n_components = min(nDmax,nClasses-1), priors = myPrior, shrinkage = None, solver = 'svd') 
    qdaMod = QDA(priors = myPrior)
    rfMod = RF()   # by default assumes equal weights

        
    # Perform CVFOLDS fold cross-validation to get performance of classifiers.
    ldaYes = 0
    qdaYes = 0
    rfYes = 0
    cvCount = 0
    
    if testInd is None:
        skf = StratifiedKFold(n_splits = cvFolds)
        skfList = skf.split(Xr, yGood)
    else:
        skfList = [(trainInd,testInd)]
    
    for train, test in skfList:
        
        # Enforce the MINCOUNT in each class for Training
        trainClasses, trainCount = np.unique(yGood[train], return_counts=True)
        goodIndClasses = np.array([n >= MINCOUNTTRAINING for n in trainCount])
        goodIndTrain = np.array([b in trainClasses[goodIndClasses] for b in yGood[train]])

        # Specity the training data set, the number of groups and priors
        yTrain = yGood[train[goodIndTrain]]
        XrTrain = Xr[train[goodIndTrain]]

        trainClasses, trainCount = np.unique(yTrain, return_counts=True) 
        ntrainClasses = trainClasses.size
        
        # Skip this cross-validation fold because of insufficient data
        if ntrainClasses < 2:
            continue
        goodInd = np.array([b in trainClasses for b in yGood[test]])    
        if (goodInd.size == 0):
            continue
           
        # Fit the data
        trainPriors = np.ones(ntrainClasses)*(1.0/ntrainClasses)
        ldaMod.priors = trainPriors
        qdaMod.priors = trainPriors
        ldaMod.fit(XrTrain, yTrain)
        qdaMod.fit(XrTrain, yTrain)        
        rfMod.fit(XrTrain, yTrain)
        
        
        ldaYes += np.around((ldaMod.score(Xr[test[goodInd]], yGood[test[goodInd]]))*goodInd.size)
        qdaYes += np.around((qdaMod.score(Xr[test[goodInd]], yGood[test[goodInd]]))*goodInd.size)
        rfYes += np.around((rfMod.score(Xr[test[goodInd]], yGood[test[goodInd]]))*goodInd.size)
        cvCount += goodInd.size


      
# Refit with all the data  for the plots
        
    ldaMod.priors = myPrior
    qdaMod.priors = myPrior
    Xrr = ldaMod.fit_transform(Xr, yGood)
    # Check labels
    for a, b in zip(classes, ldaMod.classes_):
        if a != b:
            print ('Error in ldaPlot: labels do not match')
            
# Check the within-group covariance in the rotated space 
#    covs = []
#    for group in classes:
#        Xg = Xrr[yGood == group, :]
#        covs.append(np.atleast_2d(np.cov(Xg,rowvar=False)))
#    withinCov = np.average(covs, axis=0, weights=myPrior)
  
    # Print the five largest coefficients of first 3 DFA
    MAXCOMP = 3        # Maximum number of DFA componnents
    MAXWEIGHT = 5     # Maximum number of weights printed for each componnent
    
    ncomp = min(MAXCOMP, nClasses-1)
    nweight = min(MAXWEIGHT, nD)
    
    # The scalings_ has the eigenvectors of the LDA in columns and the pca.componnents has the eigenvectors of PCA in columns
    weights = np.dot(ldaMod.scalings_[:,0:ncomp].T, pca.components_)
    
    print('LDA Weights:')
    for ic in range(ncomp):
        idmax = np.argsort(np.abs(weights[ic,:]))[::-1]
        print('DFA %d: '%ic, end = '')
        for iw in range(nweight):
            if Xcolname is None:
                colstr = 'C%d' % idmax[iw]
            else:
                colstr = Xcolname[idmax[iw]]
            print('%s %.3f; ' % (colstr, float(weights[ic, idmax[iw]]) ), end='')
        print()
        
    if plotFig:
        dimVal = 0.8    # Overall diming of background so that points can be seen
        # Obtain fits in this rotated space for display purposes   
        ldaMod.fit(Xrr, yGood)    
        qdaMod.fit(Xrr, yGood)
        rfMod.fit(Xrr, yGood)
    
        XrrMean = Xrr.mean(0)
                
        # Make a mesh for plotting
        x1, x2 = np.meshgrid(np.arange(-6.0, 6.0, 0.1), np.arange(-6.0, 6.0, 0.1))
        xm1 = np.reshape(x1, -1)
        xm2 = np.reshape(x2, -1)
        nxm = np.size(xm1)
        Xm = np.zeros((nxm, Xrr.shape[1]))
        Xm[:,0] = xm1
        if Xrr.shape[1] > 1 :
            Xm[:,1] = xm2
        
        for ix in range(2,Xrr.shape[1]):
            Xm[:,ix] = np.squeeze(np.ones((nxm,1)))*XrrMean[ix]
        
        XmcLDA = np.zeros((nxm, 4))  # RGBA values for color for LDA
        XmcQDA = np.zeros((nxm, 4))  # RGBA values for color for QDA
        XmcRF = np.zeros((nxm, 4))  # RGBA values for color for RF

    
        # Predict values on mesh for plotting based on the first two DFs     
        yPredLDA = ldaMod.predict_proba(Xm) 
        yPredQDA = qdaMod.predict_proba(Xm) 
        yPredRF = rfMod.predict_proba(Xm)

    
        # Transform the predictions in color codes
        maxLDA = yPredLDA.max()
        for ix in range(nxm) :
            cWeight = yPredLDA[ix,:]                               # Prob for all classes
            cWinner = ((cWeight == cWeight.max()).astype('float')) # Winner takes all 
            # XmcLDA[ix,:] = np.dot(cWeight, cClasses)/nClasses
            XmcLDA[ix,:] = np.dot(cWinner*cWeight, cClasses)
            XmcLDA[ix,3] = (cWeight.max()/maxLDA)*dimVal
    
        # Plot the surface of probability    
        plt.figure(facecolor='white', figsize=(10,4))
        plt.subplot(131)
        Zplot = XmcLDA.reshape(np.shape(x1)[0], np.shape(x1)[1],4)
        plt.imshow(Zplot, zorder=0, extent=[-6, 6, -6, 6], origin='lower', interpolation='none', aspect='auto')
        if nClasses > 2:
            plt.scatter(Xrr[:,0], Xrr[:,1], c=cValGood, s=40, zorder=1)
        else:
            plt.scatter(Xrr,(np.random.rand(Xrr.size)-0.5)*12.0 , c=cValGood, s=40, zorder=1) 
        plt.title('%s: LDA %d/%d' % (titleStr, ldaYes, cvCount))
        plt.axis('square')
        plt.xlim((-6, 6))
        plt.ylim((-6, 6))    
        plt.xlabel('DFA 1')
        plt.ylabel('DFA 2')

        if removeTickLabels:
            ax = plt.gca()
        
            labels = [item.get_text() for item in ax.get_xticklabels()]
            empty_string_labels = ['']*len(labels)
            ax.set_xticklabels(empty_string_labels)
            
            labels = [item.get_text() for item in ax.get_yticklabels()]
            empty_string_labels = ['']*len(labels)
            ax.set_yticklabels(empty_string_labels)
        
    
        # Transform the predictions in color codes
        maxQDA = yPredQDA.max()
        for ix in range(nxm) :
            cWeight = yPredQDA[ix,:]                               # Prob for all classes
            cWinner = ((cWeight == cWeight.max()).astype('float')) # Winner takes all 
            # XmcLDA[ix,:] = np.dot(cWeight, cClasses)/nClasses
            XmcQDA[ix,:] = np.dot(cWinner*cWeight, cClasses)
            XmcQDA[ix,3] = (cWeight.max()/maxQDA)*dimVal
    
        # Plot the surface of probability  

        plt.subplot(132)
        Zplot = XmcQDA.reshape(np.shape(x1)[0], np.shape(x1)[1],4)
        plt.imshow(Zplot, zorder=0, extent=[-6, 6, -6, 6], origin='lower', interpolation='none', aspect='auto')
        if nClasses > 2:
            plt.scatter(Xrr[:,0], Xrr[:,1], c=cValGood, s=40, zorder=1)
        else:
            plt.scatter(Xrr,(np.random.rand(Xrr.size)-0.5)*12.0 , c=cValGood, s=40, zorder=1) 
        plt.title('%s: QDA %d/%d' % (titleStr, qdaYes, cvCount))
        plt.xlabel('DFA 1')
        plt.ylabel('DFA 2')
        plt.axis('square')
        plt.xlim((-6, 6))
        plt.ylim((-6, 6))
           
        if removeTickLabels:
            ax = plt.gca()
            labels = [item.get_text() for item in ax.get_xticklabels()]
            empty_string_labels = ['']*len(labels)
            ax.set_xticklabels(empty_string_labels)
        
            labels = [item.get_text() for item in ax.get_yticklabels()]
            empty_string_labels = ['']*len(labels)
            ax.set_yticklabels(empty_string_labels)
   
        # Transform the predictions in color codes
        maxRF = yPredRF.max()
        for ix in range(nxm) :
            cWeight = yPredRF[ix,:]           # Prob for all classes
            cWinner = ((cWeight == cWeight.max()).astype('float')) # Winner takes all 
            # XmcLDA[ix,:] = np.dot(cWeight, cClasses)/nClasses  # Weighted colors does not work
            XmcRF[ix,:] = np.dot(cWinner*cWeight, cClasses)
            XmcRF[ix,3] = (cWeight.max()/maxRF)*dimVal
    
    # Plot the surface of probability    
        plt.subplot(133)
        Zplot = XmcRF.reshape(np.shape(x1)[0], np.shape(x1)[1],4)
        plt.imshow(Zplot, zorder=0, extent=[-6, 6, -6, 6], origin='lower', interpolation='none', aspect='auto')
        if nClasses > 2:    
            plt.scatter(Xrr[:,0], Xrr[:,1], c=cValGood, s=40, zorder=1)
        else:
            plt.scatter(Xrr,(np.random.rand(Xrr.size)-0.5)*12.0 , c=cValGood, s=40, zorder=1) 
            
        plt.title('%s: RF %d/%d' % (titleStr, rfYes, cvCount))
        plt.xlabel('DFA 1')
        plt.ylabel('DFA 2')
        plt.axis('square')
        plt.xlim((-6, 6))
        plt.ylim((-6, 6))
        
        if removeTickLabels:
            ax = plt.gca()
                        
            labels = [item.get_text() for item in ax.get_xticklabels()]
            empty_string_labels = ['']*len(labels)
            ax.set_xticklabels(empty_string_labels)
        
            labels = [item.get_text() for item in ax.get_yticklabels()]
            empty_string_labels = ['']*len(labels)
            ax.set_yticklabels(empty_string_labels)
        
        plt.show()
        plt.savefig('%s/%s.png' % (figdir,titleStr), format='png', dpi=1000)


    # Results
    ldaYes = int(ldaYes)
    qdaYes = int(qdaYes)
    rfYes = int(rfYes)
    
    p = 1.0/nClasses
    ldaP = 0
    qdaP = 0
    rfP = 0
    
    for k in range(ldaYes, cvCount+1):
        ldaP += binom.pmf(k, cvCount, p)
        
    for k in range(qdaYes, cvCount+1):
        qdaP += binom.pmf(k, cvCount, p)
        
    for k in range(rfYes, cvCount+1):
        rfP += binom.pmf(k, cvCount, p)
        
    print ("Number of classes %d. Chance level %.2f %%" % (nClasses, 100.0/nClasses))
    print ("%s LDA: %.2f %% (%d/%d p=%.4f)" % (titleStr, 100.0*ldaYes/cvCount, ldaYes, cvCount, ldaP))
    print ("%s QDA: %.2f %% (%d/%d p=%.4f)" % (titleStr, 100.0*qdaYes/cvCount, qdaYes, cvCount, qdaP))
    print ("%s RF: %.2f %% (%d/%d p=%.4f)" % (titleStr, 100.0*rfYes/cvCount, rfYes, cvCount, rfP))
    return ldaYes, qdaYes, rfYes, cvCount, ldaP, qdaP, rfP, nClasses, weights




from __future__ import division, print_function

import operator

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from sklearn.cross_validation import KFold
from sklearn.linear_model import Ridge


class RadialBasis1D(object):
    """ Uses kernels to project elements of an array of scalars x into a higher dimensional space of
        kernel coefficients.
    """

    def __init__(self):
        self.centers = None
        self.bw = None
        self.df = None

        self.alpha = None
        self.r2 = None
        self.w = None
        self.b = None

    def fit(self, x, df=3, verbose=True, kernel='gaussian'):
        self.df = df
        # get centers of kernels, partition space into df+1 regions
        p = 100. / (df+1)
        self.centers = np.array([np.percentile(x, int((k+1)*p)) for k in range(df)])
        assert self.centers.min() >= x.min()
        assert self.centers.max() <= x.max()
        assert len(np.unique(self.centers)) == len(self.centers), '# of unique Gaussian kernels is less than the degrees of freedom!'

        # estimate bandwidth (standard deviation) for each kernel
        self.bw = np.zeros([df])
        for k in range(df):
            if k > 0:
                start = self.centers[k-1]
            else:
                start = x.min()

            if k < df-1:
                end = self.centers[k+1]
            else:
                end = x.max()

            i = (x >= start) & (x <= end)
            assert i.sum() > 1, "too few datapoints: start=%0.6f, end=%0.6f, i.sum()=%d" % (start, end, i.sum())
            # print('start=%0.6f, end=%0.6f, i.sum()=%d, bw=%0.6f' % (start, end, i.sum(), x[i].std(ddof=1)))
            self.bw[k] = x[i].std(ddof=1) / 1.

        # compute the basis representation of x
        B = self.basis(x)

        # use ridge regression to find a good representation
        alphas = np.logspace(-2, 5, 20)
        cv_sets = list(KFold(len(x), n_folds=10))

        model_perfs = list()
        for alpha in alphas:
            perfs = list()
            for k,(training_indices,test_indices) in enumerate(cv_sets):

                Xtrain = B[training_indices, :]
                ytrain = x[training_indices]

                Xtest = B[test_indices, :]
                ytest = x[test_indices]

                rr = Ridge(alpha=alpha, fit_intercept=True)
                rr.fit(Xtrain, ytrain)
                ypred = rr.predict(Xtest)

                sst = np.sum((ytest - ytrain.mean())**2)
                sse = np.sum((ytest - ypred)**2)
                r2 = 1. - (sse / sst)

                # print('Fold %d: r2=%0.2f' % (k, r2))
                perfs.append({'r2':r2, 'w':rr.coef_, 'b':rr.intercept_})

            r2_mean = np.mean([d['r2'] for d in perfs])
            w_mean = np.mean([d['w'] for d in perfs])
            b_mean = np.mean([d['b'] for d in perfs])

            model_perfs.append({'alpha':alpha, 'r2':r2_mean, 'w':w_mean, 'b':b_mean})

        # get the best model
        model_perfs.sort(key=operator.itemgetter('r2'), reverse=True)

        self.alpha = model_perfs[0]['alpha']
        self.r2 = model_perfs[0]['r2']
        self.w = model_perfs[0]['w']
        self.b = model_perfs[0]['b']

        if verbose:
            print('RadialBasis1D: alpha=%0.6f, r2=%0.2f' % (self.alpha, self.r2))
            if self.alpha == max(alphas):
                print('RadialBasis1D WARNING: maximum alpha value encountered! Something might be wrong with your data.')
                print('x on range (%0.6f, %0.6f)' % (x.min(), x.max()))
                print('centers=')
                print(self.centers)
                print('bw=')
                print(self.bw)
                self.plot(x)
                plt.show()

    def basis(self, x):
        # compute the basis
        B = np.zeros([len(x), self.df])
        for k in range(self.df):
            B[:, k] = np.exp(-(x - self.centers[k]) ** 2 / self.bw[k])
        return B


    def transform(self, x):
        # compute the basis representation for X
        B = self.basis(x)

        # multiply the basis by coefficients, elementwise
        return B * self.w

    def plot(self, x):

        xs = sorted(x)
        X = self.transform(xs)
        xsmooth = np.linspace(x.min(), x.max(), 100)

        plt.figure()
        gs = plt.GridSpec(1, 3)

        ax = plt.subplot(gs[0, :2])
        plt.hist(x, bins=20, color='m', alpha=0.7, normed=True)
        plt.plot(x, np.zeros([len(x)]), 'k|', markersize=25)
        for k in range(self.df):
            y = np.exp(-(xsmooth - self.centers[k])**2 / self.bw[k])
            plt.plot(xsmooth, y, 'b-', alpha=0.7, linewidth=4.0)

        ax = plt.subplot(gs[0, 2:])
        absmax = np.abs(X).max()
        plt.imshow(X, interpolation='nearest', aspect='auto', cmap=plt.cm.seismic, vmin=-absmax, vmax=absmax)
        plt.colorbar()


def cubic_spline_basis(x, num_knots=3, return_knots=False, knots=None):

    if knots is None:
        p = 100. / (num_knots + 1)
        knots = np.array([np.percentile(x, int((k + 1) * p)) for k in range(num_knots)])
        assert knots.min() >= x.min()
        assert knots.max() <= x.max()
        if len(np.unique(knots)) != len(knots):
            # print('[cubic_spline_basis] number of unique kernels is less than the degrees of freedom, trying wider knot spacing (q10, q50, q90)')
            knots = [np.percentile(x, 10), np.percentile(x, 50), np.percentile(x, 90)]
        assert len(np.unique(knots)) == len(knots), '# of unique kernels is less than the degrees of freedom!'

    num_knots = len(knots)
    df = num_knots+3
    B = np.zeros([len(x), df])
    for k in range(3):
        B[:, k] = x**(k+1)

    for k in range(num_knots):
        i = x > knots[k]
        B[i, k+3] = (x[i]-knots[k])**3

    if return_knots:
        return B,knots
    return B


def natural_spline_basis(x, num_knots=3):
    p = 100. / (num_knots + 1)
    knots = np.array([np.percentile(x, int((k + 1) * p)) for k in range(num_knots)])
    assert knots.min() >= x.min()
    assert knots.max() <= x.max()
    assert len(np.unique(knots)) == len(knots), '# of unique kernels is less than the degrees of freedom!'

    df = num_knots
    B = np.zeros([len(x), df])
    B[:, 0] = x

    def _dk(_k):
        _i1 = x > knots[_k]
        _i2 = x > knots[-1]
        _x1 = np.zeros([len(x)])
        _x2 = np.zeros([len(x)])

        _x1[_i1] = x - knots[_k]
        _x2[_i2] = x - knots[-1]

        _num = (_x1**3 - _x2**3)
        _denom = (knots[-1] - knots[_k])
        assert abs(_denom) > 0, "denom=0, _k=%d, _i1.sum()=%d, _i2.sum()=%d" % (_k, _i1.sum(), _i2.sum())

        return _num / _denom

    for k in range(num_knots-2):
        B[:, k + 1] = _dk(k) - _dk(num_knots-2)

    return B

    
